import collections
import time

from six.moves import cPickle, range

from smqtk.representation.descriptor_index import DescriptorIndex

# Try to import required module
try:
    import solr
except ImportError:
    solr = None


__author__ = "paul.tunison@kitware.com"


class SolrDescriptorIndex (DescriptorIndex):
    """
    Descriptor index that uses a Solr instance as a backend storage medium.

    Fields where components are stored within a document are specified at
    construction time. We optionally set the ``id`` field to a string UUID.
    ``id`` is set because it is a common, required field for unique
    identification of documents.

    Descriptor UUIDs should maintain their uniqueness when converted to a
    string, otherwise this backend will not work well when querying.

    """

    @classmethod
    def is_usable(cls):
        return solr is not None

    def __init__(self, solr_conn_addr, index_uuid,
                 index_uuid_field, d_uid_field, descriptor_field,
                 timestamp_field, solr_params=None,
                 commit_on_add=True, max_boolean_clauses=1024,
                 pickle_protocol=-1):
        """
        Construct a descriptor index pointing to a Solr instance.

        :param solr_conn_addr: HTTP(S) address for the Solr index to use
        :type solr_conn_addr: str

        :param index_uuid: Unique ID for the descriptor index to use within the
            configured Solr index.
        :type index_uuid: str

        :param index_uuid_field: Solr index field to store/locate index UUID
            value.
        :type index_uuid_field: str

        :param d_uid_field: Solr index field to store/locate descriptor UUID
            values
        :type d_uid_field: str

        :param descriptor_field: Solr index field to store the code-associated
            descriptor object.
        :type descriptor_field: str

        :param timestamp_field: Solr index field to store floating-point UNIX
            timestamps.
        :type timestamp_field: str

        :param solr_params: Dictionary of additional keyword parameters to set
            in the ``solr.Solr`` instance used. See the ``pysolr``
            documentation for available parameters and values.
        :type solr_params: dict[str, object]

        :param commit_on_add: Immediately commit changes when one or many
            descriptor are added.
        :type commit_on_add: bool

        :param max_boolean_clauses: Solr instance's configured
            maxBooleanClauses configuration property (found in solrconfig.xml
            file). This is needed so we can correctly chunk up batch queries
            without breaking the server. This may also be less than the Solr
            instance's set value.
        :type max_boolean_clauses: int

        :param pickle_protocol: Pickling protocol to use. We will use -1 by
            default (latest version, probably binary).
        :type pickle_protocol: int

        """
        super(SolrDescriptorIndex, self).__init__()

        self.index_uuid = index_uuid

        self.index_uuid_field = index_uuid_field
        self.d_uid_field = d_uid_field
        self.descriptor_field = descriptor_field
        self.timestamp_field = timestamp_field

        self.commit_on_add = commit_on_add
        self.max_boolean_clauses = int(max_boolean_clauses)
        assert self.max_boolean_clauses >= 2, "Need more clauses"

        self.pickle_protocol = pickle_protocol

        self.solr_params = solr_params
        self.solr = solr.Solr(solr_conn_addr, **solr_params)

    def __getstate__(self):
        return self.get_config()

    def __setstate__(self, state):
        state['solr'] = solr.Solr(state["solr_conn_addr"],
                                  **state['solr_params'])
        del state['solr_conn_addr']
        self.__dict__.update(state)

    def _doc_for_code_descr(self, d):
        """
        Generate standard identifying document base for the given
        descriptor element.
        """
        uuid = d.uuid()
        return {
            'id': '-'.join([self.index_uuid,  uuid]),
            self.index_uuid_field: self.index_uuid,
            self.d_uid_field: uuid,
        }

    def get_config(self):
        return {
            "solr_conn_addr": self.solr.url,
            "index_uuid": self.index_uuid,
            "index_uuid_field": self.index_uuid_field,
            "d_uid_field": self.d_uid_field,
            "descriptor_field": self.descriptor_field,
            "timestamp_field": self.timestamp_field,
            "solr_params": self.solr_params,
            "commit_on_add": self.commit_on_add,
            "max_boolean_clauses": self.max_boolean_clauses,
            "pickle_protocol": self.pickle_protocol,
        }

    def count(self):
        """
        :return: Number of descriptor elements stored in this index.
        :rtype: int
        """
        return int(self.solr.
                   select("%s:%s AND %s:*"
                          % (self.index_uuid_field, self.index_uuid,
                             self.descriptor_field))
                   .numFound)

    def clear(self):
        """
        Clear this descriptor index's entries.
        """
        self.solr.delete_query("%s:%s"
                               % (self.index_uuid_field, self.index_uuid))
        self.solr.commit()

    def has_descriptor(self, uuid):
        """
        Check if a DescriptorElement with the given UUID exists in this index.

        :param uuid: UUID to query for
        :type uuid: collections.Hashable

        :return: True if a DescriptorElement with the given UUID exists in this
            index, or False if not.
        :rtype: bool

        """
        # Try to select the descriptor
        # TODO: Probably a better way of doing this that's more efficient.
        return bool(
            self.solr.select("%s:%s AND %s:%s"
                             % (self.index_uuid_field, self.index_uuid,
                                self.d_uid_field, uuid)).numFound
        )

    def add_descriptor(self, descriptor):
        """
        Add a descriptor to this index.

        Adding the same descriptor multiple times should not add multiple copies
        of the descriptor in the index (based on UUID). Added descriptors
        overwrite indexed descriptors based on UUID.

        :param descriptor: Descriptor to index.
        :type descriptor: smqtk.representation.DescriptorElement

        """
        doc = self._doc_for_code_descr(descriptor)
        doc[self.descriptor_field] = cPickle.dumps(descriptor,
                                                   self.pickle_protocol)
        doc[self.timestamp_field] = time.time()
        self.solr.add(doc, commit=self.commit_on_add)

    def add_many_descriptors(self, descriptors):
        """
        Add multiple descriptors at one time.

        Adding the same descriptor multiple times should not add multiple copies
        of the descriptor in the index (based on UUID). Added descriptors
        overwrite indexed descriptors based on UUID.

        :param descriptors: Iterable of descriptor instances to add to this
            index.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        documents = []
        for d in descriptors:
            doc = self._doc_for_code_descr(d)
            doc[self.descriptor_field] = cPickle.dumps(d, self.pickle_protocol)
            doc[self.timestamp_field] = time.time()
            documents.append(doc)
        self.solr.add_many(documents)
        if self.commit_on_add:
            self.solr.commit()

    def get_descriptor(self, uuid):
        """
        Get the descriptor in this index that is associated with the given UUID.

        :param uuid: UUID of the DescriptorElement to get.
        :type uuid: collections.Hashable

        :raises KeyError: The given UUID doesn't associate to a
            DescriptorElement in this index.

        :return: DescriptorElement associated with the queried UUID.
        :rtype: smqtk.representation.DescriptorElement

        """
        return tuple(self.get_many_descriptors(uuid))[0]

    def get_many_descriptors(self, *uuids):
        """
        Get an iterator over descriptors associated to given descriptor UUIDs.

        :param uuids: Iterable of descriptor UUIDs to query for.
        :type uuids: collections.Hashable

        :raises KeyError: A given UUID doesn't associate with a
            DescriptorElement in this index.

        :return: Iterator of descriptors associated 1-to-1 to given uuid values.
        :rtype: collections.Iterable[smqtk.representation.DescriptorElement]

        """
        # Chunk up query based on max clauses available to us

        def batch_query(_batch):
            """
            :param _batch: Batch of UIDs to select.
            :type _batch: list[collections.Hashable]
            """
            query = ' OR '.join([self.d_uid_field + (':%s' % _uid)
                                 for _uid in _batch])
            r = self.solr.select("%s:%s AND (%s)"
                                 % (self.index_uuid_field, self.index_uuid,
                                    query))
            # result batches come in chunks of 10
            for doc in r.results:
                yield cPickle.loads(doc[self.descriptor_field])
            for j in range(r.numFound // 10):
                r = r.next_batch()
                for doc in r.results:
                    yield cPickle.loads(doc[self.descriptor_field])

        batch = []
        for uid in uuids:
            batch.append(uid)

            # Will end up using max_clauses-1 OR statements, and one AND
            if len(batch) == self.max_boolean_clauses:
                for d in batch_query(batch):
                    yield d
                batch = []

        # tail batch
        if batch:
            assert len(batch) < self.max_boolean_clauses
            for d in batch_query(batch):
                yield d

    def remove_descriptor(self, uuid):
        """
        Remove a descriptor from this index by the given UUID.

        :param uuid: UUID of the DescriptorElement to remove.
        :type uuid: collections.Hashable

        :raises KeyError: The given UUID doesn't associate to a
            DescriptorElement in this index.

        """
        self.remove_many_descriptors([uuid])

    def remove_many_descriptors(self, uuids):
        """
        Remove descriptors associated to given descriptor UUIDs from this index.

        :param uuids: Iterable of descriptor UUIDs to remove.
        :type uuids: collections.Iterable[collections.Hashable]

        :raises KeyError: A given UUID doesn't associate with a
            DescriptorElement in this index.

        """
        # Chunk up operation based on max clauses available to us

        def batch_op(_batch):
            """
            :param _batch: UIDs to remove from index.
            :type _batch: collections.Iterable[collections.Hashable]
            """
            uuid_query = ' OR '.join([self.d_uid_field + (':%s' % str(_uid))
                                      for _uid in _batch])
            self.solr.delete("%s:%s AND (%s)"
                             % (self.index_uuid_field, self.index_uuid,
                                uuid_query))

        batch = collections.deque()
        for uid in uuids:
            batch.append(uid)

            # Will end up using max_clauses-1 OR statements, and one AND
            if len(batch) == self.max_boolean_clauses:
                batch_op(batch)
            batch.clear()

        # tail batch
        if batch:
            batch_op(batch)

    def iterkeys(self):
        """
        Return an iterator over indexed descriptor keys, which are their UUIDs.
        """
        r = self.solr.select('%s:%s %s:*'
                             % (self.index_uuid_field, self.index_uuid,
                                self.d_uid_field))
        for doc in r.results:
            yield doc[self.d_uid_field]
        for _ in range(r.numFound // 10):
            r = r.next_batch()
            for doc in r.results:
                yield doc[self.d_uid_field]

    def iterdescriptors(self):
        """
        Return an iterator over indexed descriptor element instances.
        """
        r = self.solr.select('%s:%s %s:*'
                             % (self.index_uuid_field, self.index_uuid,
                                self.descriptor_field))
        for doc in r.results:
            yield cPickle.loads(doc[self.descriptor_field])
        for _ in range(r.numFound // 10):
            r = r.next_batch()
            for doc in r.results:
                yield cPickle.loads(doc[self.descriptor_field])

    def iteritems(self):
        """
        Return an iterator over indexed descriptor key and instance pairs.
        """
        r = self.solr.select('%s:%s %s:* %s:*'
                             % (self.index_uuid_field, self.index_uuid,
                                self.d_uid_field, self.descriptor_field))
        for doc in r.results:
            d = cPickle.loads(doc[self.descriptor_field])
            yield d.uuid(), d
        for _ in range(r.numFound // 10):
            r = r.next_batch()
            for doc in r.results:
                d = cPickle.loads(doc[self.descriptor_field])
                yield d.uuid(), d



"""KWCNN descriptor generator definition."""
from collections import deque
import os
import io
import itertools
import logging
import multiprocessing
import multiprocessing.pool

import numpy
import PIL.Image
import PIL.ImageFile
import six
# noinspection PyUnresolvedReferences
from six.moves import range, zip

from smqtk.algorithms.descriptor_generator import (DescriptorGenerator,
                                                   DFLT_DESCRIPTOR_FACTORY)
from smqtk.utils.bin_utils import report_progress

try:
    from six.moves import cPickle as pickle
except ImportError:
    import pickle

try:
    import kwcnn
    from .autoencoder_model_def import AutoEncoderModel
except ImportError as ex:
    kwcnn = None
    AutoEncoderModel = None


__author__ = 'jason.parham@kitware.com,paul.tunison@kitware.com'

__all__ = [
    "KWCNNDescriptorGenerator",
]


DEFAULT_MODEL_FILEPATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'kwcnnmodel.npy'
)


class KWCNNDescriptorGenerator (DescriptorGenerator):
    """KWCNN descriptor generator to compute autoencoder-generated features."""

    @classmethod
    def is_usable(cls):
        """Return a boolean for if the descriptor generator is usable."""
        valid = kwcnn is not None
        if not valid:
            cls.get_logger().debug("KWCNN python module cannot be imported")
        return valid

    def __init__(self, network_model_filepath=DEFAULT_MODEL_FILEPATH,
                 batch_size=32, use_gpu=False, gpu_device_id=0,
                 network_is_greyscale=False, load_truncated_images=False,
                 pixel_rescale=None, input_scale=None,
                 pre_initialize_network=True):
        """
        Create a KWCNN CNN descriptor generator.

        :param network_model_filepath: The path to the trained KWCNN ``.npy``
            model file to use.
        :type network_model_filepath: str

        :param batch_size: The maximum number of images to process in one feed
            forward of the network. This is especially important for GPUs since
            they can only process a batch that will fit in the GPU memory space.
        :type batch_size: int

        :param use_gpu: If KWCNN should try to use the GPU
        :type use_gpu: bool

        :param gpu_device_id: Integer ID of the GPU device to use. Only used if
            ``use_gpu`` is True.
        :type gpu_device_id: int

        :param network_is_greyscale: If the network is expecting a greyscale
            format of pixels instead of BGR.
        :type network_is_greyscale: bool

        :param load_truncated_images: If we should be lenient and force loading
            of truncated image bytes. This is False by default.
        :type load_truncated_images: bool

        :param pixel_rescale: Re-scale image pixel values before being
            transformed by kwcnn (before mean subtraction, etc)
            into the given tuple ``(min, max)`` range. By default, images are
            loaded in the ``[0, 255]`` range. Refer to the image mean being used
            for desired input pixel scale.
        :type pixel_rescale: None | (float, float)

        :param input_scale: Optional floating-point scalar value to scale values
            of kwcnn network input data AFTER mean subtraction. This value is
            directly multiplied against the pixel values.
        :type input_scale: None | float

        """
        super(KWCNNDescriptorGenerator, self).__init__()

        self.batch_size = int(batch_size)

        self.use_gpu = bool(use_gpu)
        self.gpu_device_id = int(gpu_device_id)
        self.gpu_device_tag = 'gpu%d' % (self.gpu_device_id, )

        self.network_is_greyscale = bool(network_is_greyscale)
        # assert self.network_is_greyscale is False, 'Only color model supported'  # NOQA
        self.load_truncated_images = bool(load_truncated_images)
        self.pixel_rescale = pixel_rescale
        self.input_scale = input_scale

        self.network_model_filepath = str(network_model_filepath)

        assert self.batch_size > 0, \
            "Batch size must be greater than 0 (got %d)" % self.batch_size
        assert self.gpu_device_id >= 0, \
            "GPU Device ID must be greater than 0 (got %d)" % self.gpu_device_id

        # Network setup variables
        self.data = None
        self.model = None
        self.network = None

        self._setup_network(pre_initialize_network=pre_initialize_network)

    def __getstate__(self):
        """Get the state of the descriptor generator."""
        return self.get_config()

    def __setstate__(self, state):
        """Set the state of the descriptor generator."""
        self.__dict__.update(state)
        self._setup_network()

    # noinspection PyProtectedMember
    def _setup_network(self, pre_initialize_network=True):
        """Initialize KWCNN data, model, and network objects."""
        # Check KWCNN
        try:
            assert kwcnn is not None
        except AssertionError:
            self._log.error("KWCNN python module not imported")
            raise

        # Check Theano CPU/GPU state vs. configured
        try:
            if self.use_gpu:
                self._log.debug("Using GPU")
                assert kwcnn.tpl._lasagne.USING_GPU
                assert kwcnn.tpl._lasagne.USING_DEVICE == self.gpu_device_tag
            else:
                self._log.debug("Using CPU")
                assert not kwcnn.tpl._lasagne.USING_GPU
        except AssertionError:
            self._log.error("Theano mis-configured for specified device!")
            url = "http://deeplearning.net/software/theano/library/" \
                  "config.html#environment-variables"  # NOQA
            self._log.error("Review Theano documentation here: %s" % (url, ))

            self._log.error("Requested configuration:")
            # Check the configuration requested by the SMQTK configuration
            self._log.error("\t\t Using CPU       : %s", not self.use_gpu)
            self._log.error("\t\t Using GPU       : %s", self.use_gpu)
            self._log.error("\t\t Using GPU Device: %s", self.gpu_device_tag)

            self._log.error("Theano configuration:")
            # Check the configuration reported by imported Theano
            self._log.error("\t Imported theano module configuration")
            self._log.error("\t\t Using CPU       : %s",
                            not kwcnn.tpl._lasagne.USING_GPU)
            self._log.error("\t\t Using GPU       : %s",
                            kwcnn.tpl._lasagne.USING_GPU)
            self._log.error("\t\t Using GPU Device: %s",
                            kwcnn.tpl._lasagne.USING_DEVICE)

            # Check the $HOME/.theanorc file for configuration
            self._log.error("\t $HOME/.theanorc configuration file")
            theanorc_filepath = os.path.join("~", ".theanorc")
            theanorc_filepath = os.path.expanduser(theanorc_filepath)
            if os.path.exists(theanorc_filepath):
                with open(theanorc_filepath, "r") as theanorc_file:
                    for line in theanorc_file.readlines():
                        self._log.error("\t\t %s" % (line.strip(), ))
            else:
                self._log.error("\t\t NO CONFIGURATION FILE")

            # Check the $THEANO_FLAGS environment variable for configuration
            self._log.error("\t $THEANO_FLAGS environment variable")
            theano_flags = os.environ.get("THEANO_FLAGS", "").strip()
            if len(theano_flags) > 0:
                theano_flag_list = theano_flags.split(",")
                for theano_flag in theano_flag_list:
                    self._log.error("\t\t %s" % (theano_flag.strip(), ))
            else:
                self._log.error("\t\t NO ENVIRONMENT VARIABLE")
            # Raise RuntimeError for the user to address the configuration issue
            raise RuntimeError("Theano misconfigured for specified device")

        # Create KWCNN Data, Model, and Network primitives
        self._log.debug("Initializing KWCNN Data")
        self.data = kwcnn.core.KWCNN_Data()

        # Create trimmed model, if it does not exist
        self._log.debug("Initializing KWCNN Model")

        USE_TRIMMED_NETWORK = True

        if USE_TRIMMED_NETWORK:
            trimmed_filepath = self.network_model_filepath
            trimmed_filepath = trimmed_filepath.replace('.npy', '.trimmed.npy')
            if not os.path.exists(trimmed_filepath):
                with open(self.network_model_filepath, 'rb') as model_file:
                    model_dict = pickle.load(model_file)
                key_list = ['best_weights', 'best_fit_weights']
                for key in key_list:
                    layer_list = model_dict[key]
                    model_dict[key] = layer_list[:-2]
                with open(trimmed_filepath, 'wb') as model_file:
                    pickle.dump(model_dict, model_file,
                                protocol=pickle.HIGHEST_PROTOCOL)
            self.network_model_filepath = trimmed_filepath

        # Load model
        self.model = AutoEncoderModel(self.network_model_filepath,
                                      greyscale=self.network_is_greyscale,
                                      trimmed=USE_TRIMMED_NETWORK)

        self._log.debug("Initializing KWCNN Network")
        self.network = kwcnn.core.KWCNN_Network(self.model, self.data)

        # Pre-initialize network during network setup
        if pre_initialize_network:
            # Get the input shape for the KWCNN model
            self.input_shape = self.model._input_shape()
            input_height, input_width, input_channels = self.input_shape
            # Create a temporary numpy array of empty data of correct shape
            temp_shape = (16, input_height, input_width, input_channels, )
            temp_arrays = numpy.zeros(temp_shape, dtype=numpy.float32)
            # Give the dummy data to the KWCNN data object
            self.data.set_data_list(temp_arrays, quiet=True)
            # Test with dummy data, which will compile and load the model
            self._log.debug("Building and compiling KWCNN model...")
            self.network.test(quiet=True)  # Throw away output
            self._log.debug("done")

    def get_config(self):
        """
        Get the configuration dictionary for the descriptor generator.

        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        In the common case, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict
        """
        return {
            "network_model_filepath": self.network_model_filepath,
            "batch_size": self.batch_size,
            "use_gpu": self.use_gpu,
            "gpu_device_id": self.gpu_device_id,
            "network_is_greyscale": self.network_is_greyscale,
            "load_truncated_images": self.load_truncated_images,
            "pixel_rescale": self.pixel_rescale,
            "input_scale": self.input_scale,
        }

    def valid_content_types(self):
        """
        Return the MIME types that are valid for this descriptor generator.

        :return: A set valid MIME type content types that this descriptor can
            handle.
        :rtype: set[str]
        """
        return {
            "image/tiff",
            "image/png",
            "image/jpeg",
        }

    def _compute_descriptor(self, data):
        raise NotImplementedError("Shouldn't get here as "
                                  "compute_descriptor[_async] is being "
                                  "overridden")

    def compute_descriptor(self, data, descr_factory=DFLT_DESCRIPTOR_FACTORY,
                           overwrite=False):
        """
        Given some data, return a descriptor element containing a descriptor
        vector.

        :raises RuntimeError: Descriptor extraction failure of some kind.
        :raises ValueError: Given data element content was not of a valid type
            with respect to this descriptor.

        :param data: Some kind of input data for the feature descriptor.
        :type data: smqtk.representation.DataElement

        :param descr_factory: Factory instance to produce the wrapping
            descriptor element instance. The default factory produces
            ``DescriptorMemoryElement`` instances by default.
        :type descr_factory: smqtk.representation.DescriptorElementFactory

        :param overwrite: Whether or not to force re-computation of a descriptor
            vector for the given data even when there exists a precomputed
            vector in the generated DescriptorElement as generated from the
            provided factory. This will overwrite the persistently stored vector
            if the provided factory produces a DescriptorElement implementation
            with such storage.
        :type overwrite: bool

        :return: Result descriptor element. UUID of this output descriptor is
            the same as the UUID of the input data element.
        :rtype: smqtk.representation.DescriptorElement

        """
        m = self.compute_descriptor_async([data], descr_factory, overwrite,
                                          procs=1)
        return m[data.uuid()]

    def compute_descriptor_async(self, data_iter,
                                 descr_factory=DFLT_DESCRIPTOR_FACTORY,
                                 overwrite=False, procs=None, **kwds):
        """
        Asynchronously compute feature data for multiple data items.

        :param data_iter: Iterable of data elements to compute features for.
            These must have UIDs assigned for feature association in return
            value.
        :type data_iter: collections.Iterable[smqtk.representation.DataElement]

        :param descr_factory: Factory instance to produce the wrapping
            descriptor element instance. The default factory produces
            ``DescriptorMemoryElement`` instances by default.
        :type descr_factory: smqtk.representation.DescriptorElementFactory

        :param overwrite: Whether or not to force re-computation of a descriptor
            vectors for the given data even when there exists precomputed
            vectors in the generated DescriptorElements as generated from the
            provided factory. This will overwrite the persistently stored
            vectors if the provided factory produces a DescriptorElement
            implementation such storage.
        :type overwrite: bool

        :param procs: Optional specification of how many processors to use
            when pooling sub-tasks. If None, we attempt to use all available
            cores.
        :type procs: int | None

        :raises ValueError: An input DataElement was of a content type that we
            cannot handle.

        :return: Mapping of input DataElement UUIDs to the computed descriptor
            element for that data. DescriptorElement UUID's are congruent with
            the UUID of the data element it is the descriptor of.
        :rtype: dict[collections.Hashable,
                     smqtk.representation.DescriptorElement]

        """
        # Create DescriptorElement instances for each data elem.
        #: :type: dict[collections.Hashable, smqtk.representation.DataElement]
        data_elements = {}
        #: :type: dict[collections.Hashable, smqtk.representation.DescriptorElement]
        descr_elements = {}
        self._log.debug("Checking content types; aggregating data/descriptor "
                        "elements.")
        prog_rep_state = [0] * 7
        for data in data_iter:
            ct = data.content_type()
            if ct not in self.valid_content_types():
                self._log.error("Cannot compute descriptor from content type "
                                "'%s' data: %s)" % (ct, data))
                raise ValueError("Cannot compute descriptor from content type "
                                 "'%s' data: %s)" % (ct, data))
            data_elements[data.uuid()] = data
            descr_elements[data.uuid()] = \
                descr_factory.new_descriptor(self.name, data.uuid())
            report_progress(self._log.debug, prog_rep_state, 1.0)
        self._log.debug("Given %d unique data elements", len(data_elements))

        # Reduce procs down to the number of elements to process if its smaller
        if len(data_elements) < (procs or multiprocessing.cpu_count()):
            procs = len(data_elements)
        if procs == 0:
            raise ValueError("No data elements provided")

        # For thread safely, only use .append() and .popleft() (queue)
        uuid4proc = deque()

        def check_get_uuid(descriptor_elem):
            if overwrite or not descriptor_elem.has_vector():
                # noinspection PyUnresolvedReferences
                uuid4proc.append(descriptor_elem.uuid())

        # Using thread-pool due to in-line function + updating local deque
        p = multiprocessing.pool.ThreadPool(procs)
        try:
            p.map(check_get_uuid, six.itervalues(descr_elements))
        finally:
            p.close()
            p.join()
        del p
        self._log.debug("%d descriptors already computed",
                        len(data_elements) - len(uuid4proc))

        if uuid4proc:
            self._log.debug("Converting deque to tuple for segmentation")
            uuid4proc = tuple(uuid4proc)

            # Split UUIDs into groups equal to our batch size, and an option
            # tail group that is less than our batch size.
            tail_size = len(uuid4proc) % self.batch_size
            batch_groups = (len(uuid4proc) - tail_size) // self.batch_size
            self._log.debug("Processing %d batches of size %d", batch_groups,
                            self.batch_size)
            if tail_size:
                self._log.debug("Processing tail group of size %d", tail_size)

            if batch_groups:
                for g in range(batch_groups):
                    self._log.debug("Starting batch: %d of %d",
                                    g + 1, batch_groups)
                    batch_uuids = \
                        uuid4proc[g * self.batch_size:(g + 1) * self.batch_size]
                    self._process_batch(batch_uuids, data_elements,
                                        descr_elements, procs)

            if tail_size:
                batch_uuids = uuid4proc[-tail_size:]
                self._log.debug("Starting tail batch (size=%d)",
                                len(batch_uuids))
                self._process_batch(batch_uuids, data_elements, descr_elements,
                                    procs)

        self._log.debug("forming output dict")
        return dict((data_elements[k].uuid(), descr_elements[k])
                    for k in data_elements)

    def _process_batch(self, uuids4proc, data_elements, descr_elements, procs):
        """
        Run a number of data elements through the network in batches.

        Compute results of elements based on the number of UUIDs given,
        returning the vectors of descriptors

        :param uuids4proc: UUIDs of the source data to run in the network as a
            batch.
        :type uuids4proc: collections.Sequence[collections.Hashable]

        :param data_elements: Mapping of UUID to data element for input data.
        :type data_elements: dict[collections.Hashable,
                                  smqtk.representation.DataElement]

        :param descr_elements: Mapping of UUID to descriptor element based on
            input data elements.
        :type descr_elements: dict[collections.Hashable,
                                   smqtk.representation.DescriptorElement]

        :param procs: The number of asynchronous processes to run for loading
            images. This may be None to just use all available cores.
        :type procs: None | int

        """
        self._log.debug("Loading image pixel arrays")
        uid_num = len(uuids4proc)
        p = multiprocessing.Pool(procs)
        img_arrays = p.map(
            _process_load_img_array,
            zip(
                (data_elements[uid] for uid in uuids4proc),
                itertools.repeat(self.network_is_greyscale, uid_num),
                itertools.repeat(self.input_shape, uid_num),
                itertools.repeat(self.load_truncated_images, uid_num),
                itertools.repeat(self.pixel_rescale, uid_num),
            )
        )
        p.close()
        p.join()

        self._log.debug("Loading image numpy array into KWCNN Data object")
        self.data.set_data_list(img_arrays, quiet=True)

        self._log.debug("Performing forward inference using KWCNN Network")
        test_results = self.network.test(quiet=True)
        descriptor_list = test_results['probability_list']

        self._log.debug("transform network output into descriptors")
        for uid, v in zip(uuids4proc, descriptor_list):
            if v.ndim > 1:
                # In case kwcnn generates multidimensional array (rows, 1, 1)
                descr_elements[uid].set_vector(numpy.ravel(v))
            else:
                descr_elements[uid].set_vector(v)


def _process_load_img_array(data_element, network_is_greyscale, input_shape,
                            load_truncated_images, pixel_rescale):
    """
    Helper function for multiprocessing image data loading.

    :param network_is_greyscale: If PIL should load the images in greyscale.

    :param load_truncated_images: If PIL should be allowed to load truncated
        image data. If false, and exception will be raised when encountering
        such imagery.

    :param pixel_rescale: A post-load tuple specifying a (min, max) scaling
        applied to the pixel values

    :return: Pre-processed numpy array.

    """
    PIL.ImageFile.LOAD_TRUNCATED_IMAGES = load_truncated_images
    img = PIL.Image.open(io.BytesIO(data_element.get_bytes()))
    mode = "L" if network_is_greyscale else "RGB"
    if img.mode != mode:
        img = img.convert(mode)
    # KWCNN natively uses uint8 or float32 types
    input_size = input_shape[:2]
    if img.size != input_size:
        img = img.resize(input_size, PIL.Image.LANCZOS)
    try:
        # This can fail if the image is truncated and we're not allowing the
        # loading of those images
        img_a = numpy.array(img, dtype=numpy.float32)
        img_a = img_a.reshape(input_shape)
    except Exception:
        logging.getLogger(__name__).error(
            "Failed array-ifying data element. Image may be truncated: %s",
            data_element
        )
        raise
    # Convert to BGR
    if not network_is_greyscale:
        img_a = img_a[:, :, ::-1]
    message = "Loaded invalid greyscale image with shape %s" % (img_a.shape, )
    assert img_a.ndim == 3 and img_a.shape == input_shape, message
    if pixel_rescale:
        pmin, pmax = min(pixel_rescale), max(pixel_rescale)
        r = pmax - pmin
        img_a = (img_a / (255. / r)) + pmin
    return img_a



"""
Interface for generic element-wise nearest-neighbor computation.
"""

import abc

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.utils import check_empty_iterable


class NearestNeighborsIndex (SmqtkAlgorithm):
    """
    Common interface for descriptor-based nearest-neighbor computation over a
    built index of descriptors.

    Implementations, if they allow persistent storage of their index, should
    take the necessary parameters at construction time. Persistent storage
    content should be (over)written ``build_index`` is called.

    Implementations should be thread safe and appropriately protect
    internal model components from concurrent access and modification.

    """

    def __len__(self):
        return self.count()

    @staticmethod
    def _empty_iterable_exception():
        """
        Create the exception instance to be thrown when no descriptors are
        provided to ``build_index``/``update_index``.

        :return: ValueError instance to be thrown.
        :rtype: ValueError

        """
        return ValueError("No DescriptorElement instances in provided "
                          "iterable.")

    def build_index(self, descriptors):
        """
        Build the index with the given descriptor data elements.

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :raises ValueError: No data available in the given iterable.

        :param descriptors: Iterable of descriptor elements to build index
            over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        check_empty_iterable(descriptors, self._build_index,
                             self._empty_iterable_exception())

    def update_index(self, descriptors):
        """
        Additively update the current index with the one or more descriptor
        elements given.

        If no index exists yet, a new one should be created using the given
        descriptors.

        :raises ValueError: No data available in the given iterable.

        :param descriptors: Iterable of descriptor elements to add to this
            index.
        :type descriptors: collections.Iterable[smqtk.representation
                                                     .DescriptorElement]

        """
        check_empty_iterable(descriptors, self._update_index,
                             self._empty_iterable_exception())

    def remove_from_index(self, uids):
        """
        Partially remove descriptors from this index associated with the given
        UIDs.

        :param uids: Iterable of UIDs of descriptors to remove from this index.
        :type uids: collections.Iterable[collections.Hashable]

        :raises ValueError: No data available in the given iterable.
        :raises KeyError: One or more UIDs provided do not match any stored
            descriptors.  The index should not be modified.

        """
        check_empty_iterable(uids, self._remove_from_index,
                             self._empty_iterable_exception())

    def nn(self, d, n=1):
        """
        Return the nearest `N` neighbors to the given descriptor element.

        :raises ValueError: Input query descriptor ``d`` has no vector set.
        :raises ValueError: Current index is empty.

        :param d: Descriptor element to compute the neighbors of.
        :type d: smqtk.representation.DescriptorElement

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N DescriptorElement instances, and a tuple of
            the distance values to those neighbors.
        :rtype: (tuple[smqtk.representation.DescriptorElement], tuple[float])

        """
        if not d.has_vector():
            raise ValueError("Query descriptor did not have a vector set!")
        elif not self.count():
            raise ValueError("No index currently set to query from!")
        return self._nn(d, n)

    @abc.abstractmethod
    def count(self):
        """
        :return: Number of elements in this index.
        :rtype: int
        """

    @abc.abstractmethod
    def _build_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to build the index with
        the given descriptor data elements.

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :param descriptors: Iterable of descriptor elements to build index
            over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """

    @abc.abstractmethod
    def _update_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more descriptor elements given.

        If no index exists yet, a new one should be created using the given
        descriptors.

        :param descriptors: Iterable of descriptor elements to add to this
            index.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """

    @abc.abstractmethod
    def _remove_from_index(self, uids):
        """
        Internal method to be implemented by sub-classes to partially remove
        descriptors from this index associated with the given UIDs.

        :param uids: Iterable of UIDs of descriptors to remove from this index.
        :type uids: collections.Iterable[collections.Hashable]

        :raises KeyError: One or more UIDs provided do not match any stored
            descriptors.

        """

    @abc.abstractmethod
    def _nn(self, d, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbors to the given descriptor element.

        When this internal method is called, we have already checked that there
        is a vector in ``d`` and our index is not empty.

        :param d: Descriptor element to compute the neighbors of.
        :type d: smqtk.representation.DescriptorElement

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N DescriptorElement instances, and a tuple of
            the distance values to those neighbors.
        :rtype: (tuple[smqtk.representation.DescriptorElement], tuple[float])

        """



"""
Train a supervised classifier based on an IQR session state dump.

Descriptors used in IQR, and thus referenced via their UUIDs in the IQR session
state dump, must exist external to the IQR web-app (uses a non-memory backend).
This is needed so that this script might access them for classifier training.

Getting an IQR Session's State Information
==========================================
Click the "Save IQR State" button to download the IqrState file encapsulating
the descriptors of positively and negatively marked items. These descriptors
will be used to train the configured SupervisedClassifier.

"""

import logging
import os

from smqtk.algorithms import SupervisedClassifier
from smqtk.algorithms import get_classifier_impls

from smqtk.iqr import IqrSession

from smqtk.representation import DescriptorElementFactory
from smqtk.representation.descriptor_element.local_elements \
    import DescriptorMemoryElement

from smqtk.utils.bin_utils import (
    basic_cli_parser,
    utility_main_helper,
)
from smqtk.utils.plugin import make_config
from smqtk.utils.plugin import from_plugin_config


def get_cli_parser():
    parser = basic_cli_parser(__doc__)
    parser.add_argument('-i', '--iqr-state',
                        help="Path to the ZIP file saved from an IQR session.")
    return parser


def get_default_config():
    return {
        "classifier": make_config(get_classifier_impls()),
    }


def train_classifier_iqr(config, iqr_state_fp):
    #: :type: smqtk.algorithms.SupervisedClassifier
    classifier = from_plugin_config(
        config['classifier'],
        get_classifier_impls(sub_interface=SupervisedClassifier)
    )

    # Load state into an empty IqrSession instance.
    with open(iqr_state_fp, 'rb') as f:
        state_bytes = f.read().strip()
    descr_factory = DescriptorElementFactory(DescriptorMemoryElement, {})
    iqrs = IqrSession()
    iqrs.set_state_bytes(state_bytes, descr_factory)

    # Positive descriptor examples for training are composed of those from
    # external and internal sets. Same for negative descriptor examples.
    pos = iqrs.positive_descriptors | iqrs.external_positive_descriptors
    neg = iqrs.negative_descriptors | iqrs.external_negative_descriptors
    classifier.train(class_examples={'positive': pos, 'negative': neg})


def main():
    args = get_cli_parser().parse_args()
    config = utility_main_helper(get_default_config, args)

    log = logging.getLogger(__name__)
    log.debug("Showing debug messages.")

    iqr_state_fp = args.iqr_state

    if not os.path.isfile(iqr_state_fp):
        log.error("IQR Session info JSON filepath was invalid")
        exit(102)

    train_classifier_iqr(config, iqr_state_fp)


if __name__ == "__main__":
    main()



from __future__ import division, print_function
import mock
import unittest

from smqtk.representation import DataSet


class DummyDataSet (DataSet):

    @classmethod
    def is_usable(cls):
        return True

    def __init__(self):
        super(DummyDataSet, self).__init__()

    def __iter__(self):
        pass

    def count(self):
        pass

    def uuids(self):
        pass

    def has_uuid(self, uuid):
        pass

    def add_data(self, *elems):
        pass

    def get_data(self, uuid):
        pass

    def get_config(self):
        return {}


class TestDataSetAbstract (unittest.TestCase):

    def test_len(self):
        expected_len = 134623456

        ds = DummyDataSet()
        ds.count = mock.MagicMock(return_value=expected_len)

        self.assertEqual(len(ds), expected_len)

    def test_getitem_mock(self):
        expected_key = 'foo'
        expected_value = 'bar'

        def expected_effect(k):
            if k == expected_key:
                return expected_value
            raise RuntimeError("not expected key")

        ds = DummyDataSet()
        ds.get_data = mock.MagicMock(side_effect=expected_effect)

        self.assertRaisesRegexp(
            RuntimeError,
            "^not expected key$",
            ds.__getitem__, 'unexpectedKey'
        )
        self.assertEqual(ds[expected_key], expected_value)

    def test_contains(self):
        # Contains built-in hook expects data element and requests UUID from
        # that.
        expected_uuid = 'some uuid'

        mock_data_element = mock.MagicMock()
        mock_data_element.uuid = mock.MagicMock(return_value=expected_uuid)

        def expected_has_uuid_effect(k):
            if k == expected_uuid:
                return True
            return False

        ds = DummyDataSet()
        ds.has_uuid = mock.MagicMock(side_effect=expected_has_uuid_effect)

        self.assertTrue(mock_data_element in ds)
        ds.has_uuid.assert_called_once_with(expected_uuid)

        mock_data_element.uuid.return_value = 'not expected uuid'
        self.assertFalse(mock_data_element in ds)



from smqtk.utils import merge_dict
from smqtk.utils import plugin
from smqtk.representation import (
    DataElement,
    DataSet,
    KeyValueStore,
    get_key_value_store_impls
)
from smqtk.representation.key_value.memory import MemoryKeyValueStore


DFLT_KVSTORE = MemoryKeyValueStore()


class KVSDataSet (DataSet):
    """
    DataSet backed by a KeyValueStore implementation.

    Since KeyValue stores should be able to contain arbitrary hashable keys and
    arbitrary values, this leverages available implementations for the
    KeyValueStore interface.
    """

    @classmethod
    def is_usable(cls):
        """
        This implementation is always usable.
        :return: True
        :rtype: bool
        """
        # No external dependencies.
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict
        """
        c = super(KVSDataSet, cls).get_default_config()
        c['kvstore'] = merge_dict(
            plugin.make_config(get_key_value_store_impls()),
            plugin.to_plugin_config(c['kvstore'])
        )
        return c

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: KVSDataSet
        """
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        # Convert KVStore config to instance for constructor.
        kvs_inst = plugin.from_plugin_config(config_dict['kvstore'],
                                             get_key_value_store_impls())
        config_dict['kvstore'] = kvs_inst

        return super(KVSDataSet, cls).from_config(config_dict, False)

    def __init__(self, kvstore=DFLT_KVSTORE):
        """
        Create new instance.

        If no key-value store is provided, and empty in-memory implementation
        instance is used.

        :param kvstore: Backing key-value store instance.
        :type kvstore: smqtk.representation.KeyValueStore
        """
        super(KVSDataSet, self).__init__()
        assert isinstance(kvstore, KeyValueStore), \
            "Not constructed with a KeyValueStore instance."
        self._kvstore = kvstore

    def get_config(self):
        return {
            'kvstore': plugin.to_plugin_config(self._kvstore)
        }

    def __iter__(self):
        """
        :return: Generator over the DataElements contained in this set in no
            particular order.
        """
        for key in self._kvstore.keys():
            yield key

    def count(self):
        """
        :return: The number of data elements in this set.
        :rtype: int
        """
        return len(self._kvstore)

    def uuids(self):
        """
        :return: A new set of uuids represented in this data set.
        :rtype: set
        """
        return set(self._kvstore.keys())

    def has_uuid(self, uuid):
        """
        Test if the given uuid refers to an element in this data set.

        :param uuid: Unique ID to test for inclusion. This should match the
            type that the set implementation expects or cares about.
        :type uuid: collections.Hashable

        :return: True if the given uuid matches an element in this set, or
            False if it does not.
        :rtype: bool
        """
        return self._kvstore.has(uuid)

    def add_data(self, *elems):
        """
        Add the given data element(s) instance to this data set.

        :param elems: Data element(s) to add
        :type elems: smqtk.representation.DataElement
        """
        d = {}
        for e in elems:
            if isinstance(e, DataElement):
                d[e.uuid()] = e
            else:
                raise ValueError("Invalid element '%s'" % e)
        self._kvstore.add_many(d)

    def get_data(self, uuid):
        """
        Get the data element the given uuid references, or raise an
        exception if the uuid does not reference any element in this set.

        :raises KeyError: If the given uuid does not refer to an element in
            this data set.

        :param uuid: The uuid of the element to retrieve.
        :type uuid: collections.Hashable

        :return: The data element instance for the given uuid.
        :rtype: smqtk.representation.DataElement
        """
        return self._kvstore.get(uuid)



import math

import numpy

# noinspection PyUnresolvedReferences
from six.moves import range

from . import ncr

try:
    # noinspection PyUnresolvedReferences
    from numba import jit
except (ImportError, TypeError):
    # Create passthrough function if numba is not installed.
    def jit(func_or_sig):
        import types
        if isinstance(func_or_sig, (types.FunctionType, types.MethodType)):
            return func_or_sig
        else:
            return lambda *args, **kwds: func_or_sig


def next_perm(v):
    """
    Compute the lexicographically next bit permutation

    Generates next permutation with a given amount of set bits,
    given the previous lexicographical value.

    Taken from http://graphics.stanford.edu/~seander/bithacks.html

    """
    t = (v | (v - 1)) + 1
    w = t | ((((t & -t) / (v & -v)) >> 1) - 1)
    return w


def iter_perms(l, n):
    """
    Return an iterator over bit combinations of length ``l`` with ``n`` set
    bits.

    :raises StopIteration: If ``n`` <= 0 or normal completion.

    :param l: Total bit length to work with. The ``n`` in nCr problem.
    :type l: int

    :param n: Number of bits to be set in permutations. The ``r`` in nCr
        problem.
    :type n: int

    :return: List of bit vector permutations of the value ``(1<<n)-1`` over
        ``l`` bits.
    :rtype: list[int]

    """
    if n <= 0:
        raise StopIteration()
    n = min(l, n)
    s = (1 << n) - 1
    yield s
    for _ in range(ncr(l, n) - 1):
        s = next_perm(s)
        yield s


def neighbor_codes(b, c, d):
    """
    Iterate through integers of bit length ``b``, where ``b`` is the number
    of bits, that are ``d`` hamming distance away from query code ``c``.

    This will yield a number of elements equal to ``nCr(b, d)``.

    We expect ``d`` to be the integer hamming distance,
    e.g. h(001101, 100101) == 2, not 0.333.

    :param b: integer bit length
    :param b: int

    :param c: Query small-code integer
    :type c: int

    :param d: Integer hamming distance
    :type d: int

    """
    if not d:
        yield c
    else:
        for fltr in iter_perms(b, d):
            yield c ^ fltr


@jit
def bit_vector_to_int(v):
    """
    Transform a numpy vector representing a sequence of binary bits [0 | >0]
    into an integer representation.

    This version handles vectors of up to 64bits in size.

    :param v: 1D Vector of bits
    :type v: numpy.ndarray

    :return: Integer equivalent
    :rtype: int

    """
    c = 0
    for b in v:
        c = (c << 1) + int(b)
    return c


def bit_vector_to_int_large(v):
    """
    Transform a numpy vector representing a sequence of binary bits [0 | >0]
    into an integer representation.

    This function is the special form that can handle very large integers
    (>64bit).

    :param v: 1D Vector of bits
    :type v: numpy.ndarray

    :return: Integer equivalent
    :rtype: int

    """
    c = 0
    for b in v:
        c = (c << 1) + int(b)
    return c


@jit
def int_to_bit_vector(integer, bits=0):
    """
    Transform integer into a bit vector, optionally of a specific length.

    This version handles vectors of up to 64bits in size.

    :raises ValueError: If ``bits`` specified is smaller than the required bits
        to represent the given ``integer`` value.

    :param integer: integer to convert
    :type integer: int

    :param bits: Optional fixed number of bits that should be represented by the
        vector.
    :type bits: Optional specification of the size of returned vector.

    :return: Bit vector as numpy array (big endian).
    :rtype: numpy.ndarray[bool]

    """
    # Can't use math version because floating-point precision runs out after
    # about 2^48
    # -2 to remove length of '0b' string prefix
    size = len(bin(integer)) - 2

    if bits and (bits - size) < 0:
        raise ValueError("%d bits too small to represent integer value %d."
                         % (bits, integer))

    # Converting integer to array
    v = numpy.zeros(bits or size, numpy.bool_)
    for i in range(0, size):
        v[-(i+1)] = integer & 1
        integer >>= 1

    return v


def int_to_bit_vector_large(integer, bits=0):
    """
    Transform integer into a bit vector, optionally of a specific length.

    This function is the special form that can handle very large integers
    (>64bit).

    :raises ValueError: If ``bits`` specified is smaller than the required bits
        to represent the given ``integer`` value.

    :param integer: integer to convert
    :type integer: int

    :param bits: Optional fixed number of bits that should be represented by the
        vector.
    :type bits: Optional specification of the size of returned vector.

    :return: Bit vector as numpy array (big endian).
    :rtype: numpy.ndarray[bool]

    """
    # Can't use math version because floating-point precision runs out after
    # about 2^48
    # -2 to remove length of '0b' string prefix
    size = len(bin(integer)) - 2

    if bits and (bits - size) < 0:
        raise ValueError("%d bits too small to represent integer value %d."
                         % (bits, integer))

    # Converting integer to array
    v = numpy.zeros(bits or size, numpy.bool_)
    for i in range(0, size):
        v[-(i+1)] = integer & 1
        integer >>= 1

    return v


def popcount(v):
    """
    Count the number of bits set (number of 1-bits, not 0-bits).

    Pure python popcount algorithm adapted implementation at
    https://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel.

    Maximum known stable value that can be passed through this method:
    2**256 - 2. See the ``popcount.v_max`` function property.

    :param v: Integer to count the set bits of. Must be a 32-bit integer or
        less.
    :type v: int

    :return: Number of set bits in the given integer ``v``.
    :rtype: int

    """
    # TODO: C implementation of this
    #       since this version, being in python, isn't faster than counting 1's
    #       in result of ``bin`` function.

    # Cannot take the log of 0.
    if not v:
        return 0

    # T is the number of bits used to represent v to the nearest power of 2
    ceil, log = math.ceil, math.log
    tp = max(8, int(2**ceil(log(v.bit_length()) / log(2))))
    t = 2**tp-1
    b = tp // 8

    # bit-length constrained
    h55 = t//3
    h33 = t//15*3
    h0f = t//255*15
    h01 = t//255

    # noinspection PyAugmentAssignment
    v = v - ((v >> 1) & h55)
    v = (v & h33) + ((v >> 2) & h33)
    v = (v + (v >> 4)) & h0f
    # Need the extra ``& t`` after the multiplication in order to simulate bit
    # truncation as if v were only a tp-bit integer
    # Magic 8 represents bits ina byte
    return ((v * h01) & t) >> ((b-1) * 8)


# Maximum known stable value that can be passed as ``v``.
popcount.v_max = (2**256) - 2



"""
Interface and plugin getter for LSH algorithm hash generation functors.
"""
import abc
import os

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.utils import plugin


class LshFunctor (SmqtkAlgorithm):
    """
    Locality-sensitive hashing functor interface.

    The aim of such a function is to be able to generate hash codes
    (bit-vectors) such that similar items map to the same or similar hashes
    with a high probability. In other words, it aims to maximize hash collision
    for similar items.

    **Building Models**

    Some hash functions want to build a model based on some training set of
    descriptors. Due to the non-standard nature of algorithm training and model
    building, please refer to the specific implementation for further
    information on whether model training is needed and how it is accomplished.

    """

    def __call__(self, descriptor):
        return self.get_hash(descriptor)

    @abc.abstractmethod
    def get_hash(self, descriptor):
        """
        Get the locality-sensitive hash code for the input descriptor.

        :param descriptor: Descriptor vector we should generate the hash of.
        :type descriptor: numpy.ndarray[float]

        :return: Generated bit-vector as a numpy array of booleans.
        :rtype: numpy.ndarray[bool]

        """


def get_lsh_functor_impls(reload_modules=False):
    """
    Discover and return discovered ``LshFunctor`` classes. Keys in the returned
    map are the names of the discovered classes, and the paired values are the
    actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that
          begin with an alphanumeric character),
        - python modules listed in the environment variable
          :envvar:`LSH_FUNCTOR_PATH`
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH separator
              character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``LSH_FUNCTOR_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type :class:`.LshFunctor`
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "LSH_FUNCTOR_PATH"
    helper_var = "LSH_FUNCTOR_CLASS"
    return plugin.get_plugins(__name__, this_dir, env_var, helper_var,
                              LshFunctor, reload_modules=reload_modules)



from __future__ import division, print_function
import mock
import random
import unittest

import numpy

from smqtk.algorithms import get_nn_index_impls
from smqtk.algorithms.nn_index.flann import FlannNearestNeighborsIndex
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement


# Don't bother running tests of the class is not usable
if FlannNearestNeighborsIndex.is_usable():

    class TestFlannIndex (unittest.TestCase):

        RAND_SEED = 42

        def _make_inst(self, dist_method):
            """
            Make an instance of FlannNearestNeighborsIndex
            """
            return FlannNearestNeighborsIndex(distance_method=dist_method,
                                              random_seed=self.RAND_SEED)

        def test_impl_findable(self):
            # Already here because the implementation is reporting itself as
            # usable.
            self.assertIn(FlannNearestNeighborsIndex.__name__,
                          get_nn_index_impls())

        def test_configuration(self):
            index_filepath = '/index_filepath'
            para_filepath = '/param_fp'
            descr_cache_fp = '/descrcachefp'

            # Make configuration based on default
            c = FlannNearestNeighborsIndex.get_default_config()
            c['index_uri'] = index_filepath
            c['parameters_uri'] = para_filepath
            c['descriptor_cache_uri'] = descr_cache_fp
            c['distance_method'] = 'hik'
            c['random_seed'] = 42

            # Build based on configuration
            #: :type: FlannNearestNeighborsIndex
            index = FlannNearestNeighborsIndex.from_config(c)
            self.assertEqual(index._index_uri, index_filepath)
            self.assertEqual(index._index_param_uri, para_filepath)
            self.assertEqual(index._descr_cache_uri, descr_cache_fp)

            c2 = index.get_config()
            self.assertEqual(c, c2)

        def test_has_model_data_no_uris(self):
            f = FlannNearestNeighborsIndex()
            self.assertFalse(f._has_model_data())

        def test_has_model_data_empty_elements(self):
            f = FlannNearestNeighborsIndex('', '', '')
            self.assertFalse(f._has_model_data())

        def test_load_flann_model_empty_data_elements(self):
            # Construct index with valid, but empty, data URIs instances
            empty_data = 'base64://'
            f = FlannNearestNeighborsIndex(empty_data, empty_data, empty_data)
            # Load method should do nothing but set PID since given data was
            # empty.
            f._load_flann_model()
            self.assertIsNone(f._descr_cache)
            self.assertIsNone(f._flann)
            self.assertIsNone(f._flann_build_params)
            self.assertIsNotNone(f._pid)

        @mock.patch("smqtk.algorithms.nn_index.flann"
                    ".FlannNearestNeighborsIndex._load_flann_model")
        def test_has_model_data_valid_uris(self, _m_flann_lfm):
            # Mocking flann data loading that occurs in constructor when given
            # non-empty URI targets
            f = FlannNearestNeighborsIndex(
                'base64://bW9kZWxEYXRh',  # 'modelData'
                'base64://cGFyYW1EYXRh',  # 'paramData'
                'base64://ZGVzY3JEYXRh',  # 'descrData'
            )
            self.assertTrue(f._has_model_data())

        def test_build_index_one(self):
            d = DescriptorMemoryElement('test', 0)
            d.set_vector(numpy.zeros(8, float))
            index = self._make_inst('euclidean')
            index.build_index([d])
            self.assertListEqual(
                index._descr_cache,
                [d]
            )
            self.assertIsNotNone(index._flann)
            self.assertIsInstance(index._flann_build_params, dict)

        def test_build_index_with_cache(self):
            # Empty memory data elements for storage
            empty_data = 'base64://'
            f = FlannNearestNeighborsIndex(empty_data, empty_data, empty_data)
            # Internal elements should initialize have zero-length byte values
            self.assertEqual(len(f._index_elem.get_bytes()), 0)
            self.assertEqual(len(f._index_param_elem.get_bytes()), 0)
            self.assertEqual(len(f._descr_cache_elem.get_bytes()), 0)

            # Make unit vectors, one for each feature dimension.
            dim = 8
            test_descriptors = []
            for i in range(dim):
                v = numpy.zeros(dim, float)
                v[i] = 1.
                d = DescriptorMemoryElement('unit', i)
                d.set_vector(v)
                test_descriptors.append(d)

            f.build_index(test_descriptors)

            # Internal elements should not have non-zero byte values.
            self.assertGreater(len(f._index_elem.get_bytes()), 0)
            self.assertGreater(len(f._index_param_elem.get_bytes()), 0)
            self.assertGreater(len(f._descr_cache_elem.get_bytes()), 0)

        def test_update_index(self):
            # Build index with one descriptor, then "update" with a second
            # different descriptor checking that the new cache contains both.
            d1 = DescriptorMemoryElement('test', 0)
            d1.set_vector(numpy.zeros(8))
            d2 = DescriptorMemoryElement('test', 1)
            d2.set_vector(numpy.ones(8))

            index = self._make_inst('euclidean')
            index.build_index([d1])
            self.assertEqual(index.count(), 1)
            self.assertSetEqual(set(index._descr_cache), {d1})

            index.update_index([d2])
            self.assertEqual(index.count(), 2)
            self.assertSetEqual(set(index._descr_cache), {d1, d2})

        def test_nn_known_descriptors_euclidean_unit(self):
            dim = 5

            ###
            # Unit vectors -- Equal distance
            #
            index = self._make_inst('euclidean')
            test_descriptors = []
            for i in range(dim):
                v = numpy.zeros(dim, float)
                v[i] = 1.
                d = DescriptorMemoryElement('unit', i)
                d.set_vector(v)
                test_descriptors.append(d)
            index.build_index(test_descriptors)
            # query descriptor -- zero vector
            # -> all modeled descriptors should be equally distance (unit
            #    corners)
            q = DescriptorMemoryElement('query', 0)
            q.set_vector(numpy.zeros(dim, float))
            r, dists = index.nn(q, dim)
            # All dists should be 1.0, r order doesn't matter
            for d in dists:
                self.assertEqual(d, 1.)

        def test_nn_known_descriptors_euclidean_ordered(self):
            index = self._make_inst('euclidean')

            # make vectors to return in a known euclidean distance order
            i = 10
            test_descriptors = []
            for j in range(i):
                d = DescriptorMemoryElement('ordered', j)
                d.set_vector(numpy.array([j, j*2], float))
                test_descriptors.append(d)
            random.shuffle(test_descriptors)
            index.build_index(test_descriptors)

            # Since descriptors were build in increasing distance from (0,0),
            # returned descriptors for a query of [0,0] should be in index
            # order.
            q = DescriptorMemoryElement('query', 99)
            q.set_vector(numpy.array([0, 0], float))
            r, dists = index.nn(q, i)
            for j, d, dist in zip(range(i), r, dists):
                self.assertEqual(d.uuid(), j)
                numpy.testing.assert_equal(d.vector(), [j, j*2])

        def test_nn_known_descriptors_hik_unit(self):
            dim = 5

            ###
            # Unit vectors - Equal distance
            #
            index = self._make_inst('hik')
            test_descriptors = []
            for i in range(dim):
                v = numpy.zeros(dim, float)
                v[i] = 1.
                d = DescriptorMemoryElement('unit', i)
                d.set_vector(v)
                test_descriptors.append(d)
            index.build_index(test_descriptors)
            # query with zero vector
            # -> all modeled descriptors have no intersection, dists should be
            #    1.0, or maximum distance by histogram intersection.
            q = DescriptorMemoryElement('query', 0)
            q.set_vector(numpy.zeros(dim, float))
            r, dists = index.nn(q, dim)
            # All dists should be 1.0, r order doesn't matter
            for d in dists:
                self.assertEqual(d, 1.)

            # query with index element
            q = test_descriptors[3]
            r, dists = index.nn(q, 1)
            self.assertEqual(r[0], q)
            self.assertEqual(dists[0], 0.)

            r, dists = index.nn(q, dim)
            self.assertEqual(r[0], q)
            self.assertEqual(dists[0], 0.)

        def test_configuration(self):
            index_filepath = '/index_filepath'
            para_filepath = '/param_fp'
            descr_cache_fp = '/descrcachefp'

            # Make configuration based on default
            c = FlannNearestNeighborsIndex.get_default_config()
            c['index_uri'] = index_filepath
            c['parameters_uri'] = para_filepath
            c['descriptor_cache_uri'] = descr_cache_fp
            c['distance_method'] = 'hik'
            c['random_seed'] = 42

            # Build based on configuration
            #: :type: FlannNearestNeighborsIndex
            index = FlannNearestNeighborsIndex.from_config(c)
            self.assertEqual(index._index_uri, index_filepath)
            self.assertEqual(index._index_param_uri, para_filepath)
            self.assertEqual(index._descr_cache_uri, descr_cache_fp)

            c2 = index.get_config()
            self.assertEqual(c, c2)

        def test_build_index_no_descriptors(self):
            f = FlannNearestNeighborsIndex()
            self.assertRaises(
                ValueError,
                f.build_index, []
            )

        def test_build_index(self):
            # Empty memory data elements for storage
            empty_data = 'base64://'
            f = FlannNearestNeighborsIndex(empty_data, empty_data, empty_data)
            # Internal elements should initialize have zero-length byte values
            self.assertEqual(len(f._index_elem.get_bytes()), 0)
            self.assertEqual(len(f._index_param_elem.get_bytes()), 0)
            self.assertEqual(len(f._descr_cache_elem.get_bytes()), 0)

            # Make unit vectors, one for each feature
            dim = 8
            test_descriptors = []
            for i in range(dim):
                v = numpy.zeros(dim, float)
                v[i] = 1.
                d = DescriptorMemoryElement('unit', i)
                d.set_vector(v)
                test_descriptors.append(d)

            f.build_index(test_descriptors)

            # Internal elements should not have non-zero byte values.
            self.assertGreater(len(f._index_elem.get_bytes()), 0)
            self.assertGreater(len(f._index_param_elem.get_bytes()), 0)
            self.assertGreater(len(f._descr_cache_elem.get_bytes()), 0)



from __future__ import division, print_function
import random
import unittest

from smqtk.utils import factors


# First 100 known primes
p100 = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61,
        67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137,
        139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199,
        211,  223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271, 277,
        281, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 353, 359,
        367, 373, 379, 383, 389, 397, 401, 409, 419, 421, 431, 433, 439,
        443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509, 521,
        523, 541]


class TestSoe (unittest.TestCase):
    """
    Unit tests for caching sieve of eratosthenes function.
    """

    def setUp(self):
        """
        Clear SoE caches before each test.
        """
        # Values copied from head of factors.py file, simulating initial import.
        factors._soe_prime_cache = [2, 3]
        factors._soe_not_prime_map = {9: 3}
        factors._soe_c = 5

    def test_0(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(0),
            []
        )

    def test_1(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(1),
            []
        )

    def test_2(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(2),
            [2]
        )

    def test_3(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(3),
            [2, 3]
        )

    def test_4(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(4),
            [2, 3]
        )

    def test_5(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(5),
            [2, 3, 5]
        )

    def test_7(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(7),
            [2, 3, 5, 7]
        )

    def test_15(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(15),
            [2, 3, 5, 7, 11, 13]
        )

    def test_25(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(25),
            [2, 3, 5, 7, 11, 13, 17, 19, 23]
        )

    def test_60(self):
        self.assertListEqual(
            factors.sieve_of_eratosthenes(60),
            [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59]
        )

    def test_100_primes(self):
        self.assertListEqual(factors.sieve_of_eratosthenes(541), p100)

    def test_100_with_caching(self):
        # First ask for a value with 99 prime returns, then a value with 100 to
        # see that we did get that new extra value.
        self.assertListEqual(factors.sieve_of_eratosthenes(540), p100[:-1])
        self.assertListEqual(factors.sieve_of_eratosthenes(541), p100)

    def test_100_random_with_reset(self):
        # Randomly test for prime values <= than 100th prime, resetting the
        # cache each time.
        for _ in range(100):
            r = random.randint(0, max(p100)+1)
            expected_primes = [p for p in p100 if p <= r]
            self.setUp()  # Reset
            actual_primes = factors.sieve_of_eratosthenes(r)
            self.assertListEqual(expected_primes, actual_primes,
                                 "Unexpected return for query value: %d" % r)

    def test_100_random_with_caching(self):
        # Randomly test for prime values <= than 100th prime, NOT resetting the
        # cache each time.
        for _ in range(100):
            r = random.randint(0, max(p100) + 1)
            expected_primes = [p for p in p100 if p <= r]
            actual_primes = factors.sieve_of_eratosthenes(r)
            self.assertListEqual(expected_primes, actual_primes,
                                 "Unexpected return for query value: %d" % r)


class TestPrimeFactors (unittest.TestCase):

    def test_2(self):
        # If the N requested is itself a prime value
        self.assertListEqual(
            factors.prime_factors(2),
            [2]
        )

    def test_8(self):
        self.assertListEqual(
            factors.prime_factors(8),
            [2, 2, 2]
        )

    def test_26345(self):
        # Tests when trailing "remaining" value is not 1 because a prime was
        # greater than the square-root of N.
        self.assertListEqual(
            factors.prime_factors(26345),
            [5, 11, 479]
        )

    def test_600851475143(self):
        # Some large number with weird prime factors.  Known values taken from
        # wolfram alpha result (assuming WA is correct).
        self.assertListEqual(
            factors.prime_factors(600851475143),
            [71, 839, 1471, 6857]
        )


class TestFactors (unittest.TestCase):

    def test_600851475143(self):
        self.assertSetEqual(
            factors.factors(600851475143),
            {1, 71, 839, 1471, 6857, 59569, 104441, 486847, 1234169, 5753023,
             10086647, 87625999, 408464633, 716151937, 8462696833, 600851475143}
        )


class TestFactorPairs (unittest.TestCase):

    def test_600851475143(self):
        self.assertListEqual(
            factors.factor_pairs(600851475143),
            [(1, 600851475143),
             (71, 8462696833),
             (839, 716151937),
             (1471, 408464633),
             (6857, 87625999),
             (59569, 10086647),
             (104441, 5753023),
             (486847, 1234169)]
        )



import multiprocessing

import numpy

from smqtk.representation import DescriptorElement
from smqtk.utils.postgres import norm_psql_cmd_string, PsqlConnectionHelper

# Try to import required modules
try:
    import psycopg2
except ImportError:
    psycopg2 = None


PSQL_TABLE_CREATE_RLOCK = multiprocessing.RLock()


# noinspection SqlNoDataSourceInspection
class PostgresDescriptorElement (DescriptorElement):
    """
    Descriptor element whose vector is stored in a Postgres database.

    We assume we will work with a Postgres version of at least 9.4 (due to
    versions tested).

    Efficient connection pooling may be achieved via external utilities like
    PGBounder.

    """

    ARRAY_DTYPE = numpy.float64

    UPSERT_TABLE_TMPL = norm_psql_cmd_string("""
        CREATE TABLE IF NOT EXISTS {table_name:s} (
          {type_col:s} TEXT NOT NULL,
          {uuid_col:s} TEXT NOT NULL,
          {binary_col:s} BYTEA NOT NULL,
          PRIMARY KEY ({type_col:s}, {uuid_col:s})
        );
    """)

    SELECT_TMPL = norm_psql_cmd_string("""
        SELECT {binary_col:s}
          FROM {table_name:s}
          WHERE {type_col:s} = %(type_val)s
            AND {uuid_col:s} = %(uuid_val)s
        ;
    """)

    UPSERT_TMPL = norm_psql_cmd_string("""
        WITH upsert AS (
          UPDATE {table_name:s}
            SET {binary_col:s} = %(binary_val)s
            WHERE {type_col:s} = %(type_val)s
              AND {uuid_col:s} = %(uuid_val)s
            RETURNING *
          )
        INSERT INTO {table_name:s} ({type_col:s}, {uuid_col:s}, {binary_col:s})
          SELECT %(type_val)s, %(uuid_val)s, %(binary_val)s
            WHERE NOT EXISTS (SELECT * FROM upsert);
    """)

    @classmethod
    def is_usable(cls):
        if psycopg2 is None:
            cls.get_logger().warning("Not usable. Requires psycopg2 module")
            return False
        return True

    def __init__(self, type_str, uuid,
                 table_name='descriptors',
                 uuid_col='uid', type_col='type_str', binary_col='vector',
                 db_name='postgres', db_host=None, db_port=None, db_user=None,
                 db_pass=None, create_table=True):
        """
        Initialize new PostgresDescriptorElement attached to some database
        credentials.

        We require that storage tables treat uuid AND type string columns as
        primary keys. The type and uuid columns should be of the 'text' type.
        The binary column should be of the 'bytea' type.

        Default argument values assume a local PostgreSQL database with a table
        created via the
        ``etc/smqtk/postgres/descriptor_element/example_table_init.sql``
        file (relative to the SMQTK source tree or install root).

        NOTES:
            - Not all uuid types used here are necessarily of the ``uuid.UUID``
              type, thus the recommendation to use a ``text`` type for the
              column. For certain specific use cases they may be proper
              ``uuid.UUID`` instances or strings, but this cannot be generally
              assumed.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: Unique ID reference of the descriptor.
        :type uuid: collections.Hashable

        :param table_name: String label of the database table to use.
        :type table_name: str

        :param uuid_col: The column label for descriptor UUID storage
        :type uuid_col: str

        :param type_col: The column label for descriptor type string storage.
        :type type_col: str

        :param binary_col: The column label for descriptor vector binary
            storage.
        :type binary_col: str

        :param db_host: Host address of the Postgres server. If None, we
            assume the server is on the local machine and use the UNIX socket.
            This might be a required field on Windows machines (not tested yet).
        :type db_host: str | None

        :param db_port: Port the Postgres server is exposed on. If None, we
            assume the default port (5423).
        :type db_port: int | None

        :param db_name: The name of the database to connect to.
        :type db_name: str

        :param db_user: Postgres user to connect as. If None, postgres
            defaults to using the current accessing user account name on the
            operating system.
        :type db_user: str | None

        :param db_pass: Password for the user we're connecting as. This may be
            None if no password is to be used.
        :type db_pass: str | None

        :param create_table: If this instance should try to create the storing
            table before actions are performed against it. If the configured
            user does not have sufficient permissions to create the table and it
            does not currently exist, an exception will be raised.
        :type create_table: bool

        """
        super(PostgresDescriptorElement, self).__init__(type_str, uuid)

        self.table_name = table_name
        self.uuid_col = uuid_col
        self.type_col = type_col
        self.binary_col = binary_col
        self.create_table = create_table

        self.db_name = db_name
        self.db_host = db_host
        self.db_port = db_port
        self.db_user = db_user
        self.db_pass = db_pass

        self._psql_helper = None

    def __getstate__(self):
        """
        Construct serialization state.

        Due to the psql_helper containing a lock, it cannot be serialized.  This
        is OK due to our creation of the helper on demand.  The cost incurred by
        discarding the instance upon serialization is that once deserialized
        elsewhere the helper instance will have to be created.  Since this
        creation post-deserialization only happens once, this is acceptable.

        """
        state = super(PostgresDescriptorElement, self).__getstate__()
        state.update({
            "table_name": self.table_name,
            "uuid_col": self.uuid_col,
            "type_col": self.type_col,
            "binary_col": self.binary_col,
            "create_table": self.create_table,
            "db_name": self.db_name,
            "db_host": self.db_host,
            "db_port": self.db_port,
            "db_user": self.db_user,
            "db_pass": self.db_pass,
        })
        return state

    def __setstate__(self, state):
        # Base DescriptorElement parts
        super(PostgresDescriptorElement, self).__setstate__(state)
        # Our parts
        self.table_name = state['table_name']
        self.uuid_col = state['uuid_col']
        self.type_col = state['type_col']
        self.binary_col = state['binary_col']
        self.create_table = state['create_table']
        self.db_name = state['db_name']
        self.db_host = state['db_host']
        self.db_port = state['db_port']
        self.db_user = state['db_user']
        self.db_pass = state['db_pass']
        self._psql_helper = None

    def _get_psql_helper(self):
        """
        Internal method to create on demand the PSQL connection helper class.
        :return: PsqlConnectionHelper utility.
        :rtype: PsqlConnectionHelper
        """
        # `hasattr` check used for backwards compatibility when interacting with
        # databases containing elements serialized before the inclusion of this
        # helper class.
        if self._psql_helper is None:
            # Only using a transport iteration size of 1 since this element is
            # only meant to refer to a single entry in the associated table.
            self._psql_helper = PsqlConnectionHelper(
                self.db_name, self.db_host, self.db_port, self.db_user,
                self.db_pass, itersize=1,
                table_upsert_lock=PSQL_TABLE_CREATE_RLOCK
            )
            # Register table upsert command
            if self.create_table:
                self._psql_helper.set_table_upsert_sql(
                    self.UPSERT_TABLE_TMPL.format(
                        table_name=self.table_name,
                        type_col=self.type_col,
                        uuid_col=self.uuid_col,
                        binary_col=self.binary_col,
                    )
                )
        return self._psql_helper

    def get_config(self):
        return {
            "table_name": self.table_name,
            "uuid_col": self.uuid_col,
            "type_col": self.type_col,
            "binary_col": self.binary_col,
            "create_table": self.create_table,

            "db_name": self.db_name,
            "db_host": self.db_host,
            "db_port": self.db_port,
            "db_user": self.db_user,
            "db_pass": self.db_pass,
        }

    def has_vector(self):
        """
        Check if the target database has a vector for our keys.

        This also returns True if we have a cached vector since there must have
        been a source vector to draw from if there is a cache of it.

        If a vector is cached, this resets the cache expiry timeout.

        :return: Whether or not this container current has a descriptor vector
            stored.
        :rtype: bool

        """
        # Very similar to vector query, but replacing vector binary return with
        # a true/null return. Save a little bit of time compared to testing
        # vector return.
        # OLD: return self.vector() is not None

        # Using static value 'true' for binary "column" to reduce data return
        # volume.
        q_select = self.SELECT_TMPL.format(**{
            'binary_col': 'true',
            'table_name': self.table_name,
            'type_col': self.type_col,
            'uuid_col': self.uuid_col,
        })
        q_select_values = {
            "type_val": self.type(),
            "uuid_val": str(self.uuid())
        }

        def cb(cursor):
            cursor.execute(q_select, q_select_values)

        # Should either yield one or zero rows.
        psql_helper = self._get_psql_helper()
        return bool(list(psql_helper.single_execute(
            cb, yield_result_rows=True
        )))

    def vector(self):
        """
        Return this element's vector, or None if we don't have one.

        :return: Get the stored descriptor vector as a numpy array. This returns
            None of there is no vector stored in this container.
        :rtype: numpy.ndarray or None

        """
        q_select = self.SELECT_TMPL.format(**{
            "binary_col": self.binary_col,
            "table_name": self.table_name,
            "type_col": self.type_col,
            "uuid_col": self.uuid_col,
        })
        q_select_values = {
            "type_val": self.type(),
            "uuid_val": str(self.uuid())
        }

        # query execution callback
        # noinspection PyProtectedMember
        def cb(cursor):
            # type: (psycopg2._psycopg.cursor) -> None
            cursor.execute(q_select, q_select_values)

        # This should only fetch a single row.  Cannot yield more than one due
        # use of primary keys.
        psql_helper = self._get_psql_helper()
        r = list(psql_helper.single_execute(cb, yield_result_rows=True))
        if not r:
            return None
        else:
            b = r[0][0]
            v = numpy.frombuffer(b, self.ARRAY_DTYPE)
            return v

    def set_vector(self, new_vec):
        """
        Set the contained vector.

        If this container already stores a descriptor vector, this will
        overwrite it.

        If we are configured to use caching, and one has not been cached yet,
        then we cache the vector and start a thread to monitor access times and
        to remove the cache if the access timeout has expired.

        If a vector was already cached, this new vector replaces the old one,
        the vector database-side is replaced, and the cache expiry timeout is
        reset.

        :raises ValueError: ``new_vec`` was not a numpy ndarray.

        :param new_vec: New vector to contain. This must be a numpy array.
        :type new_vec: numpy.ndarray

        :returns: Self.
        :rtype: PostgresDescriptorElement

        """
        if not isinstance(new_vec, numpy.ndarray):
            new_vec = numpy.copy(new_vec)

        if new_vec.dtype != self.ARRAY_DTYPE:
            try:
                new_vec = new_vec.astype(self.ARRAY_DTYPE)
            except TypeError:
                raise ValueError("Could not convert input to a vector of type "
                                 "%s." % self.ARRAY_DTYPE)

        q_upsert = self.UPSERT_TMPL.strip().format(**{
            "table_name": self.table_name,
            "binary_col": self.binary_col,
            "type_col": self.type_col,
            "uuid_col": self.uuid_col,
        })
        q_upsert_values = {
            "binary_val": psycopg2.Binary(new_vec),
            "type_val": self.type(),
            "uuid_val": str(self.uuid()),
        }

        # query execution callback
        # noinspection PyProtectedMember
        def cb(cursor):
            # type: (psycopg2._psycopg.cursor) -> None
            cursor.execute(q_upsert, q_upsert_values)

        # No return but need to force iteration.
        psql_helper = self._get_psql_helper()
        list(psql_helper.single_execute(cb, yield_result_rows=False))
        return self



"""
Validate a list of images returning the filepaths and UUIDs of only the
valid images, or optionally, only the invalid images.
"""
from __future__ import print_function
import logging
import os
import sys

from six.moves import map

from smqtk.utils.bin_utils import (
    basic_cli_parser,
    initialize_logging,
)
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.utils.image_utils import is_valid_element
from smqtk.utils import parallel


__author__ = 'dan.lamanna@kitware.com'


def get_cli_parser():
    parser = basic_cli_parser(__doc__, configuration_group=False)

    parser.add_argument('-i', '--invert',
                        default=False, action='store_true',
                        help='Invert results, showing only invalid images.')

    g_required = parser.add_argument_group("Required Arguments")
    g_required.add_argument('-f', '--file-list',
                            type=str, default=None, metavar='PATH',
                            help='Path to a file that lists data file paths.')
    return parser


def main():
    # Print help and exit if no arguments were passed
    if len(sys.argv) == 1:
        get_cli_parser().print_help()
        sys.exit(1)

    args = get_cli_parser().parse_args()
    llevel = logging.INFO if not args.verbose else logging.DEBUG
    initialize_logging(logging.getLogger('smqtk'), llevel)
    initialize_logging(logging.getLogger('__main__'), llevel)

    log = logging.getLogger(__name__)
    log.debug('Showing debug messages.')

    if args.file_list is not None and not os.path.exists(args.file_list):
        log.error('Invalid file list path: %s', args.file_list)
        exit(103)

    def check_image(image_path):
        if not os.path.exists(image_path):
            log.warn('Invalid image path given (does not exist): %s',
                     image_path)
            return False, False
        else:
            d = DataFileElement(image_path)
            return is_valid_element(d, check_image=True), d

    with open(args.file_list) as infile:
        checked_images = parallel.parallel_map(check_image,
                                               map(str.strip, infile),
                                               name='check-image-validity',
                                               use_multiprocessing=True)

        for is_valid, dfe in checked_images:
            if dfe:  # in the case of a non-existent file
                if (is_valid and not args.invert) or \
                        (not is_valid and args.invert):
                    # We know the callback above is creating DataFileElement
                    # instances.
                    # noinspection PyProtectedMember
                    print('%s,%s' % (dfe._filepath, dfe.uuid()))


if __name__ == '__main__':
    main()



import os.path as osp

from six.moves import cPickle

from smqtk.representation.classification_element import ClassificationElement

from smqtk.exceptions import NoClassificationError
from smqtk.utils import file_utils
from smqtk.utils.string_utils import partition_string


__author__ = "paul.tunison@kitware.com"


class FileClassificationElement (ClassificationElement):

    @classmethod
    def is_usable(cls):
        return True

    def __init__(self, type_name, uuid, save_dir, subdir_split=None,
                 pickle_protocol=-1):
        """
        Initialize a file-base descriptor element.

        :param type_name: Type of classification. This is usually the name of
            the classifier that generated this result.
        :type type_name: str

        :param uuid: uuid for this classification
        :type uuid: collections.Hashable

        :param save_dir: Directory to save this element's contents. If this path
            is relative, we interpret as relative to the current working
            directory.
        :type save_dir: str | unicode

        :param subdir_split: If a positive integer, this will cause us to store
            the vector file in a subdirectory under the ``save_dir`` that was
            specified. The integer value specifies the number of splits that we
            will make in the stringification of this descriptor's UUID. If there
            happen to be dashes in this stringification, we will remove them
            (as would happen if given an uuid.UUID instance as the uuid
            element).
        :type subdir_split: None | int

        :param pickle_protocol: Pickling protocol to use. We will use -1 by
            default (latest version, probably binary).
        :type pickle_protocol: int

        """
        super(FileClassificationElement, self).__init__(type_name, uuid)

        self.save_dir = osp.abspath(osp.expanduser(save_dir))
        self.pickle_protocol = pickle_protocol

        # Saving components
        self.subdir_split = subdir_split
        if subdir_split and int(subdir_split) > 0:
            self.subdir_split = subdir_split = int(subdir_split)
            # Using all but the last split segment. This is so we don't create
            # a whole bunch of directories with a single element in them.
            save_dir = osp.join(self.save_dir,
                                *partition_string(str(uuid).replace('-', ''),
                                                  subdir_split)[:subdir_split-1]
                                )
        else:
            save_dir = self.save_dir

        self.filepath = osp.join(save_dir,
                                 "%s.%s.classification.pickle"
                                 % (self.type_name, str(self.uuid)))

    def get_config(self):
        return {
            "save_dir": self.save_dir,
            'subdir_split': self.subdir_split,
            "pickle_protocol": self.pickle_protocol,
        }

    def has_classifications(self):
        """
        :return: If this element has classification information set.
        :rtype: bool
        """
        return osp.isfile(self.filepath)

    def get_classification(self):
        """
        Get classification result map, returning a label-to-confidence dict.

        We do no place any guarantees on label value types as they may be
        represented in various forms (integers, strings, etc.).

        Confidence values are in the [0,1] range.

        :raises NoClassificationError: No classification labels/confidences yet
            set.

        :return: Label-to-confidence dictionary.
        :rtype: dict[collections.Hashable, float]

        """
        if not self.has_classifications():
            raise NoClassificationError("No classification values.")
        with open(self.filepath) as f:
            return cPickle.load(f)

    def set_classification(self, m=None, **kwds):
        """
        Set the whole classification map for this element. This will strictly
        overwrite the entire label-confidence mapping (vs. updating it)

        Label/confidence values may either be provided via keyword arguments or
        by providing a dictionary mapping labels to confidence values.

        :param m: New labels-to-confidence mapping to set.
        :type m: dict[collections.Hashable, float]

        :raises ValueError: The given label-confidence map was empty.

        """
        m = super(FileClassificationElement, self)\
            .set_classification(m, **kwds)
        file_utils.safe_create_dir(osp.dirname(self.filepath))
        with open(self.filepath, 'w') as f:
            cPickle.dump(m, f, self.pickle_protocol)



from __future__ import division, print_function
import mock
import unittest

from smqtk.algorithms.classifier import Classifier, SupervisedClassifier, \
    get_classifier_impls
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement


class DummyClassifier (Classifier):

    EXPECTED_LABELS = ['constant']

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        return {}

    def get_labels(self):
        return self.EXPECTED_LABELS

    def _classify(self, d):
        """
        Returns one "label" that is the UUID of the descriptor, mapping to the
        descriptor's vector as a list.
        """
        return {d.uuid(): d.vector().tolist()}


class TestClassifierAbstractClass (unittest.TestCase):

    def test_classify(self):
        d = DescriptorMemoryElement('test', 0)
        d.set_vector([1, 2, 3])

        c = DummyClassifier()
        e = c.classify(d)
        self.assertEqual(e.get_classification(), {
            0: [1, 2, 3]
        })
        self.assertEqual(e.uuid, d.uuid())

    def test_classify_empty_descriptor(self):
        d = DescriptorMemoryElement('test', 0)
        c = DummyClassifier()
        self.assertRaisesRegexp(
            ValueError,
            "does not contain a vector to classify",
            c.classify, d
        )

    @mock.patch('smqtk.algorithms.classifier._defaults'
                '.MemoryClassificationElement.has_classifications')
    def test_classify_no_overwrite(self, m_ce_hc):
        # Testing logic when classifier element for descriptor already has
        # stored results and we are NOT overwriting. Mocking classification
        # element used in association with default element factory.
        m_ce_hc.return_value = True

        c = DummyClassifier()
        # Record if underlying classify call invoked
        c._classify = mock.MagicMock(side_effect=c._classify)

        d = DescriptorMemoryElement('test', 0)
        d.set_vector([1, 2, 3])

        c.classify(d, overwrite=False)
        c._classify.assert_not_called()

    @mock.patch('smqtk.algorithms.classifier._defaults'
                '.MemoryClassificationElement.has_classifications')
    def test_classify_with_overwrite(self, m_ce_hc):
        # Testing logic when classification element for descriptor already has
        # stored results but we call WITH overwrite on. Mocking classification
        # element used in association with default element factory.
        m_ce_hc.return_value = True

        c = DummyClassifier()
        # Record if underlying classify call invoked
        c._classify = mock.MagicMock(side_effect=c._classify)

        d = DescriptorMemoryElement('test', 0)
        d.set_vector([1, 2, 3])

        c.classify(d, overwrite=True)
        c._classify.assert_called_once_with(d)

    def test_classify_async(self):
        # Check that async classify calls classify on all input elements
        # correctly

        # make some descriptor elements
        d_elems = []
        for i in range(20):
            d = DescriptorMemoryElement('test', i)
            d.set_vector([i])
            d_elems.append(d)

        c = DummyClassifier()
        m = c.classify_async(d_elems)

        # Due to mocking, we expect to see a map of descriptor element to the
        # result of classification, which in this case is the UUID of the
        # element
        for d in d_elems:
            self.assertIn(d, m)
            # Check for expected classification
            self.assertEqual(m[d].get_classification(),
                             {d.uuid(): d.vector().tolist()})


class TestGetClassifierImpls (unittest.TestCase):
    # IndexLabelClassifier should always be available (no dependencies)

    def test_get_classifier_impls_subclass_interface_classifier(self):
        # Simple test that primarily makes sure the function doesn't fall down.
        m = get_classifier_impls()
        self.assertIsInstance(m, dict)
        self.assertIn("IndexLabelClassifier", m)

        # Should act the same as calling with no sub-interface
        m2 = get_classifier_impls(sub_interface=Classifier)
        self.assertIsInstance(m, dict)
        self.assertIn("IndexLabelClassifier", m)
        self.assertEqual(m, m2)

    def test_get_classifier_impls_subclass_interface_supervised(self):
        # should not return when sub_interface is set to the
        # SupervisedClassifier base-class.
        m = get_classifier_impls(sub_interface=SupervisedClassifier)
        self.assertIsInstance(m, dict)
        self.assertNotIn("IndexLabelClassifier", m)



import collections
import mock
import unittest

from smqtk.exceptions import ReadOnlyError
from smqtk.representation.key_value import KeyValueStore, NO_DEFAULT_VALUE, \
    get_key_value_store_impls
from smqtk.representation.key_value.memory import MemoryKeyValueStore


class DummyKVStore (KeyValueStore):

    TEST_READ_ONLY = True
    TEST_COUNT = 0

    # base-class requirements

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        pass

    # KVStore abc methods

    def __repr__(self):
        return super(DummyKVStore, self).__repr__()

    def count(self):
        return self.TEST_COUNT

    def keys(self):
        pass

    def is_read_only(self):
        return self.TEST_READ_ONLY

    def add(self, key, value):
        super(DummyKVStore, self).add(key, value)
        return self

    def add_many(self, d):
        super(DummyKVStore, self).add_many(d)
        return self

    def has(self, key):
        pass

    def get(self, key, default=NO_DEFAULT_VALUE):
        pass

    def remove(self, key):
        super(DummyKVStore, self).remove(key)
        return self

    def remove_many(self, keys):
        super(DummyKVStore, self).remove_many(keys)
        return self

    def clear(self):
        super(DummyKVStore, self).clear()
        return self


class TestKeyValueStoreAbstract (unittest.TestCase):

    def test_len(self):
        s = DummyKVStore()

        s.TEST_COUNT = 0
        assert len(s) == 0

        s.TEST_COUNT = 23456
        assert len(s) == 23456

    def test_repr(self):
        # Should return expected template string
        expected_repr = "<DummyKVStore %s>"
        actual_repr = repr(DummyKVStore())
        self.assertEqual(actual_repr, expected_repr)

    # noinspection PyUnresolvedReferences
    def test_value_iterator(self):
        expected_keys_values = {1, 5, 2345, 'foo'}

        s = DummyKVStore()
        s.keys = mock.MagicMock(return_value=expected_keys_values)
        s.get = mock.MagicMock(side_effect=lambda v: v)

        # Make sure keys now returns expected list.
        self.assertEqual(set(s.keys()), expected_keys_values)

        # Get initial iterator. ``keys`` should have only been called once so
        # far, and ``get`` method should not have been called yet.
        # called yet.
        v_iter = s.values()
        self.assertIsInstance(v_iter, collections.Iterable)
        self.assertEqual(s.keys.call_count, 1)
        self.assertEqual(s.get.call_count, 0)

        actual_values_list = set(v_iter)
        self.assertEqual(actual_values_list, expected_keys_values)
        # Keys should have been called one more time, and get should have been
        # called an equal number of times as there are keys.
        self.assertEqual(s.keys.call_count, 2)
        self.assertEqual(s.get.call_count, len(expected_keys_values))
        s.get.assert_any_call(1)
        s.get.assert_any_call(5)
        s.get.assert_any_call(2345)
        s.get.assert_any_call('foo')

    # noinspection PyUnresolvedReferences
    def test_contains(self):
        # Test that python ``has`` keyword and __contains__ method calls the
        # ``has`` method correctly.
        s = DummyKVStore()

        s.has = mock.MagicMock(return_value=True)
        self.assertTrue('some item' in s)
        s.has.assert_called_once_with('some item')

        s.has = mock.MagicMock(return_value=False)
        self.assertFalse('other item' in s)
        s.has.assert_called_once_with('other item')

    def test_get_item(self):
        s = DummyKVStore()
        s.get = mock.Mock(return_value='expected-value')
        ev = s['some-key']
        s.get.assert_called_once_with('some-key')
        self.assertEqual(ev, 'expected-value')

    def test_add_when_read_only(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = True

        self.assertRaises(
            ReadOnlyError,
            s.add, 'k', 'v'
        )

    def test_add_when_not_read_only(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = False
        s.add('k', 'v')
        # Integer
        s.add(0, 'some value')
        # type
        s.add(object, 'some value')
        # some object instance
        s.add(object(), 'some value')

    def test_add_many_read_only(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = True
        self.assertRaises(
            ReadOnlyError,
            s.add_many, {0: 1}
        )

    def test_add_many(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = False
        s.add_many({0: 1})

    def test_remove_read_only(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = True
        self.assertRaises(
            ReadOnlyError,
            s.remove, 0
        )

    def test_remove(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = False
        s.remove(0)

    def test_remove_many_read_only(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = True
        self.assertRaises(
            ReadOnlyError,
            s.remove_many, [0, 1]
        )

    def test_remove_many(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = False
        s.remove_many([0, 1])

    def test_clear_readonly(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = True
        self.assertRaisesRegexp(
            ReadOnlyError,
            "Cannot clear a read-only DummyKVStore instance.",
            s.clear
        )

    def test_clear(self):
        s = DummyKVStore()
        s.TEST_READ_ONLY = False
        s.clear()


def test_kvstore_impl_getter():
    # At least the in-memory implementation should always be available, so make
    # sure at least that is returned from the getter.
    d = get_key_value_store_impls()
    assert isinstance(d, dict)
    assert 'MemoryKeyValueStore' in d
    assert d['MemoryKeyValueStore'] == MemoryKeyValueStore



import io
import json
import logging
import threading
import uuid
import zipfile

import six

from smqtk.algorithms.relevancy_index import get_relevancy_index_impls
from smqtk.representation.descriptor_index.memory import MemoryDescriptorIndex
from smqtk.utils import SmqtkObject
from smqtk.utils import plugin


DFLT_REL_INDEX_CONFIG = {
    "type": "LibSvmHikRelevancyIndex",
    "LibSvmHikRelevancyIndex": {
        "descr_cache_filepath": None,
    }
}


class IqrResultsDict (dict):
    """
    Dictionary subclass for containing DescriptorElement-to-float mapping.

    We expect keys to be DescriptorElement instances and the values to be floats
    between [0,1], inclusive.

    """

    def __setitem__(self, i, v):
        super(IqrResultsDict, self).__setitem__(i, float(v))

    def update(self, other=None, **kwds):
        """
        D.update([E, ]**F) -> None. Update D from dict/iterable E and F.
        If E present and has a .keys() method, does: for k in E: D[k] = E[k]
        If E present and lacks .keys() method, does: for (k, v) in E: D[k] = v
        In either case, this is followed by: for k in F: D[k] = F[k]

        Reimplemented so as to use override __setitem__ method so values are
        known to be floats.
        """
        if hasattr(other, 'keys'):
            for k in other:
                self[k] = float(other[k])
        elif other is not None:
            for k, v in other:
                self[k] = float(v)
        for k in kwds:
            self[k] = float(kwds[k])


class IqrSession (SmqtkObject):
    """
    Encapsulation of IQR Session related data structures with a centralized lock
    for multi-thread access.

    This object is compatible with the python with-statement, so when elements
    are to be used or modified, it should be within a with-block so race
    conditions do not occur across threads/sub-processes.

    """

    @property
    def _log(self):
        return logging.getLogger(
            '.'.join((self.__module__, self.__class__.__name__)) +
            "[%s]" % self.uuid
        )

    def __init__(self, pos_seed_neighbors=500,
                 rel_index_config=DFLT_REL_INDEX_CONFIG,
                 session_uid=None):
        """
        Initialize the IQR session

        This does not initialize the working index for ranking as there are no
        known positive descriptor examples at this time.

        Adjudications
        -------------
        Adjudications are carried through between initializations. This allows
        indexed material adjudicated through-out the lifetime of the session to
        stay relevant.

        :param pos_seed_neighbors: Number of neighbors to pull from the given
            ``nn_index`` for each positive exemplar when populating the working
            index, i.e. this value determines the size of the working index for
            IQR refinement. By default, we try to get 500 neighbors.

            Since there may be partial to significant overlap of near neighbors
            as a result of nn_index queries for positive exemplars, the working
            index may contain anywhere from this value's number of entries, to
            ``N*P``, where ``N`` is this value and ``P`` is the number of
            positive examples at the time of working index initialization.
        :type pos_seed_neighbors: int

        :param rel_index_config: Plugin configuration dictionary for the
            RelevancyIndex to use for ranking user adjudications. By default we
            we use an in-memory libSVM based index using the histogram
            intersection metric.
        :type rel_index_config: dict

        :param session_uid: Optional manual specification of session UUID.
        :type session_uid: str | uuid.UUID

        """
        self.uuid = session_uid or str(uuid.uuid1()).replace('-', '')
        self.lock = threading.RLock()

        self.pos_seed_neighbors = int(pos_seed_neighbors)

        # Local descriptor index for ranking, populated by a query to the
        #   nn_index instance.
        # Added external data/descriptors not added to this index.
        self.working_index = MemoryDescriptorIndex()

        # Book-keeping set so we know what positive descriptors
        # UUIDs we've used to query the neighbor index with already.
        #: :type: set[collections.Hashable]
        self._wi_seeds_used = set()

        # Descriptor elements representing data from external sources.
        #: :type: set[smqtk.representation.DescriptorElement]
        self.external_positive_descriptors = set()
        #: :type: set[smqtk.representation.DescriptorElement]
        self.external_negative_descriptors = set()

        # Descriptor references from our index (above) that have been
        #   adjudicated.
        #: :type: set[smqtk.representation.DescriptorElement]
        self.positive_descriptors = set()
        #: :type: set[smqtk.representation.DescriptorElement]
        self.negative_descriptors = set()

        # Mapping of a DescriptorElement in our relevancy search index (not the
        #   index that the nn_index uses) to the relevancy score given the
        #   recorded positive and negative adjudications.
        # This is None before any initialization or refinement occurs.
        #: :type: None | dict[smqtk.representation.DescriptorElement, float]
        self.results = None

        #
        # Algorithm Instances [+Config]
        #
        # RelevancyIndex configuration and instance that is used for producing
        #   results.
        # This is only [re]constructed when initializing the session.
        self.rel_index_config = rel_index_config
        # This is None until session initialization happens after pos/neg
        # exemplar data has been added.
        #: :type: None | smqtk.algorithms.relevancy_index.RelevancyIndex
        self.rel_index = None

    def __enter__(self):
        """
        :rtype: IqrSession
        """
        self.lock.acquire()
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()

    def ordered_results(self):
        """
        Return a tuple of the current (id, probability) result pairs in
        order of descending probability score. If there are no results yet, None
        is returned.

        :rtype: None | tuple[(smqtk.representation.DescriptorElement, float)]

        """
        with self.lock:
            if self.results:
                return tuple(sorted(six.iteritems(self.results),
                                    key=lambda p: p[1],
                                    reverse=True))
            return None

    def external_descriptors(self, positive=(), negative=()):
        """
        Add positive/negative descriptors from external data.

        These descriptors may not be a part of our working index.

        :param positive: Iterable of descriptors from external sources to
            consider positive examples.
        :type positive:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param negative: Iterable of descriptors from external sources to
            consider negative examples.
        :type negative:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        positive = set(positive)
        negative = set(negative)
        with self.lock:
            self.external_positive_descriptors.update(positive)
            self.external_positive_descriptors.difference_update(negative)

            self.external_negative_descriptors.update(negative)
            self.external_negative_descriptors.difference_update(positive)

    def adjudicate(self, new_positives=(), new_negatives=(),
                   un_positives=(), un_negatives=()):
        """
        Update current state of working index positive and negative
        adjudications based on descriptor UUIDs.

        If the same descriptor element is listed in both new positives and
        negatives, they cancel each other out, causing that descriptor to not
        be included in the adjudication.

        The given iterables must be re-traversable. Otherwise the given
        descriptors will not be properly registered.

        :param new_positives: Descriptors of elements in our working index to
            now be considered to be positively relevant.
        :type new_positives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param new_negatives: Descriptors of elements in our working index to
            now be considered to be negatively relevant.
        :type new_negatives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param un_positives: Descriptors of elements in our working index to now
            be considered not positive any more.
        :type un_positives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param un_negatives: Descriptors of elements in our working index to now
            be considered not negative any more.
        :type un_negatives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        new_positives = set(new_positives)
        new_negatives = set(new_negatives)
        un_positives = set(un_positives)
        un_negatives = set(un_negatives)

        with self.lock:
            self.positive_descriptors.update(new_positives)
            self.positive_descriptors.difference_update(un_positives)
            self.positive_descriptors.difference_update(new_negatives)

            self.negative_descriptors.update(new_negatives)
            self.negative_descriptors.difference_update(un_negatives)
            self.negative_descriptors.difference_update(new_positives)

    def update_working_index(self, nn_index):
        """
        Initialize or update our current working index using the given
        :class:`.NearestNeighborsIndex` instance given our current positively
        labeled descriptor elements.

        We only query from the index for new positive elements since the last
        update or reset.

        :param nn_index: :class:`.NearestNeighborsIndex` to query from.
        :type nn_index: smqtk.algorithms.NearestNeighborsIndex

        :raises RuntimeError: There are no positive example descriptors in this
            session to use as a basis for querying.

        """
        pos_examples = (self.external_positive_descriptors |
                        self.positive_descriptors)
        if len(pos_examples) == 0:
            raise RuntimeError("No positive descriptors to query the neighbor "
                               "index with.")

        # Not clearing working index because this step is intended to be
        # additive.
        updated = False

        # adding to working index
        self._log.info("Building working index using %d positive examples "
                       "(%d external, %d adjudicated)",
                       len(pos_examples),
                       len(self.external_positive_descriptors),
                       len(self.positive_descriptors))
        # TODO: parallel_map and reduce with merge-dict
        for p in pos_examples:
            if p.uuid() not in self._wi_seeds_used:
                self._log.debug("Querying neighbors to: %s", p)
                self.working_index.add_many_descriptors(
                    nn_index.nn(p, n=self.pos_seed_neighbors)[0]
                )
                self._wi_seeds_used.add(p.uuid())
                updated = True

        # Make new relevancy index
        if updated:
            self._log.info("Creating new relevancy index over working index.")
            #: :type: smqtk.algorithms.relevancy_index.RelevancyIndex
            self.rel_index = plugin.from_plugin_config(
                self.rel_index_config, get_relevancy_index_impls()
            )
            self.rel_index.build_index(self.working_index.iterdescriptors())

    def refine(self):
        """ Refine current model results based on current adjudication state

        :raises RuntimeError: No working index has been initialized.
            :meth:`update_working_index` should have been called after
            adjudicating some positive examples.
        :raises RuntimeError: There are no adjudications to run on. We must
            have at least one positive adjudication.

        """
        with self.lock:
            if not self.rel_index:
                raise RuntimeError("No relevancy index yet. Must not have "
                                   "initialized session (no working index).")

            # combine pos/neg adjudications + added external data descriptors
            pos = self.positive_descriptors | self.external_positive_descriptors
            neg = self.negative_descriptors | self.external_negative_descriptors

            if not pos:
                raise RuntimeError("Did not find at least one positive "
                                   "adjudication.")

            self._log.debug("Ranking working set with %d pos and %d neg total "
                            "examples.", len(pos), len(neg))
            element_probability_map = self.rel_index.rank(pos, neg)

            if self.results is None:
                self.results = IqrResultsDict()
            self.results.update(element_probability_map)

            # Force adjudicated positives and negatives to be probability 1 and
            # 0, respectively, since we want to control where they show up in
            # our results view.
            # - Not all pos/neg descriptors may be in our working index.
            for d in pos:
                if d in self.results:
                    self.results[d] = 1.0
            for d in neg:
                if d in self.results:
                    self.results[d] = 0.0

    def reset(self):
        """ Reset the IQR Search state

        No positive adjudications, reload original feature data

        """
        with self.lock:
            self.working_index.clear()
            self._wi_seeds_used.clear()
            self.positive_descriptors.clear()
            self.negative_descriptors.clear()
            self.external_positive_descriptors.clear()
            self.external_negative_descriptors.clear()

            self.rel_index = None
            self.results = None

    ###########################################################################
    # I/O Methods

    # I/O Constants. These should not be changed.
    STATE_ZIP_COMPRESSION = zipfile.ZIP_DEFLATED
    STATE_ZIP_FILENAME = "iqr_state.json"

    def get_state_bytes(self):
        """
        Get a byte representation of the current descriptor and adjudication
        state of this session.

        This does not encode current results or the relevancy index's state, but
        these can be reproduced with this state.

        :return: State representation bytes
        :rtype: bytes

        """
        def d_set_to_list(d_set):
            # Convert set of descriptors to list of tuples:
            #   [..., (uuid, type, vector), ...]
            return [(d.uuid(), d.type(), d.vector().tolist()) for d in d_set]

        with self:
            # Convert session descriptors into basic values.
            pos_d = d_set_to_list(self.positive_descriptors)
            neg_d = d_set_to_list(self.negative_descriptors)
            ext_pos_d = d_set_to_list(self.external_positive_descriptors)
            ext_neg_d = d_set_to_list(self.external_negative_descriptors)

        z_buffer = io.BytesIO()
        z = zipfile.ZipFile(z_buffer, 'w', self.STATE_ZIP_COMPRESSION)
        z.writestr(self.STATE_ZIP_FILENAME, json.dumps({
            'pos': pos_d,
            'neg': neg_d,
            'external_pos': ext_pos_d,
            'external_neg': ext_neg_d,
        }))
        z.close()
        return z_buffer.getvalue()

    def set_state_bytes(self, b, descriptor_factory):
        """
        Set this session's state to the given byte representation, resetting
        this session in the process.

        Bytes given must have been retrieved via a previous call to
        ``get_state_bytes`` otherwise this method will fail.

        Since this state may be completely different from the current state,
        this session is reset before applying the new state. Thus, any current
        ranking results are thrown away.

        :param b: Bytes to set this session's state to.
        :type b: bytes

        :param descriptor_factory: Descriptor element factory to use when
            generating descriptor elements from extracted data.
        :type descriptor_factory: smqtk.representation.DescriptorElementFactory

        :raises ValueError: The input bytes could not be loaded due to
            incompatibility.

        """
        z_buffer = io.BytesIO(b)
        z = zipfile.ZipFile(z_buffer, 'r', self.STATE_ZIP_COMPRESSION)
        if self.STATE_ZIP_FILENAME not in z.namelist():
            raise ValueError("Invalid bytes given, did not contain expected "
                             "zipped file name.")

        # Extract expected json file object
        state = json.loads(z.read(self.STATE_ZIP_FILENAME).decode())
        del z, z_buffer

        with self:
            self.reset()

            def load_descriptor(_uid, _type_str, vec_list):
                _e = descriptor_factory.new_descriptor(_type_str, _uid)
                if _e.has_vector():
                    assert _e.vector().tolist() == vec_list, \
                        "Found existing vector for UUID '%s' but vectors did " \
                        "not match."
                else:
                    _e.set_vector(vec_list)
                return _e

            # Read in raw descriptor data from the state, convert to descriptor
            # element, then store in our descriptor sets.
            for source, target in [(state['external_pos'],
                                    self.external_positive_descriptors),
                                   (state['external_neg'],
                                    self.external_negative_descriptors),
                                   (state['pos'], self.positive_descriptors),
                                   (state['neg'], self.negative_descriptors)]:
                for uid, type_str, vector_list in source:
                    e = load_descriptor(uid, type_str, vector_list)
                    target.add(e)



import abc
from collections import deque
import hashlib
import io
import logging
import mimetypes
import os
import os.path as osp
import tempfile

import six

from smqtk.exceptions import InvalidUriError, NoUriResolutionError, \
    ReadOnlyError
from smqtk.representation import SmqtkRepresentation
from smqtk.utils import file_utils
from smqtk.utils import plugin


MIMETYPES = mimetypes.MimeTypes()


class DataElement (SmqtkRepresentation, plugin.Pluggable):
    """
    Abstract interface for a byte data container.

    The primary "value" of a ``DataElement`` is the byte content wrapped. Since
    this can technically change due to external forces, we cannot guarantee that
    an element is immutable. Thus ``DataElement`` instances are not considered
    generally hashable. Specific implementations may define a ``__hash__``
    method if that implementation reflects a data source that guarantees
    immutability.

    UUIDs should be cast-able to a string and maintain unique-ness after
    conversion.

    """

    @classmethod
    def from_uri(cls, uri):
        """
        Construct a new instance based on the given URI.

        This function may not be implemented for all DataElement types.

        :param uri: URI string to resolve into an element instance
        :type uri: str

        :raises NoUriResolutionError: This element type does not implement URI
            resolution.
        :raises smqtk.exceptions.InvalidUriError: This element type could not
            resolve the provided URI string.

        :return: New element instance of our type.
        :rtype: DataElement

        """
        raise NoUriResolutionError()

    def __init__(self):
        super(DataElement, self).__init__()
        self._temp_filepath_stack = []

    # Because we can't generally guarantee external data immutability.
    __hash__ = None

    def __del__(self):
        self.clean_temp()

    def __eq__(self, other):
        return isinstance(other, DataElement) and \
               self.get_bytes() == other.get_bytes()

    def __ne__(self, other):
        return not (self == other)

    @abc.abstractmethod
    def __repr__(self):
        return self.__class__.__name__

    def _write_new_temp(self, d):
        """
        Actually write our bytes to a new temp file
        Always creates new file.

        :param d: directory to write temp file in or None to use system default.
        :returns: path to file written

        """
        if d:
            file_utils.safe_create_dir(d)
        ext = MIMETYPES.guess_extension(self.content_type() or '')
        # Exceptions because mimetypes is apparently REALLY OLD
        if ext in {'.jpe', '.jfif'}:
            ext = '.jpg'
        fd, fp = tempfile.mkstemp(
            suffix=ext or '',
            dir=d
        )
        os.close(fd)
        with open(fp, 'wb') as f:
            f.write(self.get_bytes())
        return fp

    def _clear_no_exist(self):
        """
        Clear paths in temp stack that don't exist on the system.
        """
        no_exist_paths = deque()  # tmp list of paths to remove
        for fp in self._temp_filepath_stack:
            if not osp.isfile(fp):
                no_exist_paths.append(fp)
        for fp in no_exist_paths:
            self._temp_filepath_stack.remove(fp)

    def md5(self):
        """
        Get the MD5 checksum of this element's binary content.

        :return: MD5 hex checksum of the data content.
        :rtype: str
        """
        return hashlib.md5(self.get_bytes()).hexdigest()

    def sha1(self):
        """
        Get the SHA1 checksum of this element's binary content.

        :return: SHA1 hex checksum of the data content.
        :rtype: str
        """
        return hashlib.sha1(self.get_bytes()).hexdigest()

    def sha512(self):
        """
        Get the SHA512 checksum of this element's binary content.

        :return: SHA512 hex checksum of the data content.
        :rtype: str
        """
        return hashlib.sha512(self.get_bytes()).hexdigest()

    def write_temp(self, temp_dir=None):
        """
        Write this data's bytes to a temporary file on disk, returning the path
        to the written file, whose extension is guessed based on this data's
        content type.

        It is not guaranteed that the returned file path does not point to the
        original data, i.e. writing to the returned filepath may modify the
        original data.

        NOTE:
            The file path returned should not be explicitly removed by the user.
            Instead, the ``clean_temp()`` method should be called on this
            object.

        :param temp_dir: Optional directory to write temporary file in,
            otherwise we use the platform default temporary files directory.
            If this is an empty string, we count it the same as having provided
            None.
        :type temp_dir: None or str

        :return: Path to the temporary file
        :rtype: str

        """
        # Write a new temp file if there aren't any in the stack, or if the none
        # of the entries' base directory is the provided temp_dir (when one is
        # provided).

        # Clear out paths that don't exist.
        self._clear_no_exist()

        if temp_dir:
            abs_temp_dir = osp.abspath(osp.expanduser(temp_dir))
            # Check if dir is the base of any path in the current stack.
            for tf in self._temp_filepath_stack:
                if osp.dirname(tf) == abs_temp_dir:
                    return tf
            # nothing in stack with given base directory, create new temp file
            self._temp_filepath_stack.append(self._write_new_temp(temp_dir))

        elif not self._temp_filepath_stack:
            # write new temp file to platform specific temp directory
            self._temp_filepath_stack.append(self._write_new_temp(None))

        # return last written temp file.
        return self._temp_filepath_stack[-1]

    def clean_temp(self):
        """
        Clean any temporary files created by this element. This does nothing if
        no temporary files have been generated for this element yet.
        """
        if len(self._temp_filepath_stack):
            for fp in self._temp_filepath_stack:
                if os.path.isfile(fp):
                    os.remove(fp)
            self._temp_filepath_stack = []

    def uuid(self):
        """
        UUID for this data element.

        This many take different forms from integers to strings to a uuid.UUID
        instance. This must return a hashable data type.

        By default, this ends up being the hex stringification of the SHA1 hash
        of this data's bytes. Specific implementations may provide other UUIDs,
        however.

        :return: UUID value for this data element. This return value should be
            hashable.
        :rtype: collections.Hashable

        """
        # TODO(paul.tunison): Change to SHA512.
        return self.sha1()

    def to_buffered_reader(self):
        """
        Wrap this element's bytes in a ``io.BufferedReader`` instance for use as
        file-like object for reading.

        As we use the ``get_bytes`` function, this element's bytes must safely
        fit in memory for this method to be usable.

        :return: New BufferedReader instance
        :rtype: io.BufferedReader

        """
        return io.BufferedReader(io.BytesIO(self.get_bytes()))

    def is_read_only(self):
        """
        :return: If this element can only be read from.
        :rtype: bool
        """
        return not self.writable()

    ###
    # Abstract methods
    #

    @abc.abstractmethod
    def content_type(self):
        """
        :return: Standard type/subtype string for this data element, or None if
            the content type is unknown.
        :rtype: str or None
        """

    @abc.abstractmethod
    def is_empty(self):
        """
        Check if this element contains no bytes.

        The intend of this method is to quickly check if there is any data
        behind this element, ideally without having to read all/any of the
        underlying data.

        :return: If this element contains 0 bytes.
        :rtype: bool

        """

    @abc.abstractmethod
    def get_bytes(self):
        """
        :return: Get the bytes for this data element.
        :rtype: bytes
        """

    @abc.abstractmethod
    def writable(self):
        """
        :return: if this instance supports setting bytes.
        :rtype: bool
        """

    @abc.abstractmethod
    def set_bytes(self, b):
        """
        Set bytes to this data element.

        Not all implementations may support setting bytes (check ``writable``
        method return).

        This base abstract method should be called by sub-class implementations
        first. We check for mutability based on ``writable()`` method return and
        invalidate checksum caches.

        :param b: bytes to set.
        :type b: str

        :raises ReadOnlyError: This data element can only be read from / does
            not support writing.

        """
        if not self.writable():
            raise ReadOnlyError("This %s element is read only." % self)


def get_data_element_impls(reload_modules=False):
    """
    Discover and return discovered ``DataElement`` classes. Keys in the
    returned map are the names of the discovered classes, and the paired values
    are the actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that begin
          with an alphanumeric character),
        - python modules listed in the environment variable
          ``DATA_ELEMENT_PATH``
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH separator
              character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``DATA_ELEMENT_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type ``DataElement``
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "DATA_ELEMENT_PATH"
    helper_var = "DATA_ELEMENT_CLASS"
    return plugin.get_plugins(__name__, this_dir, env_var, helper_var,
                              DataElement, reload_modules=reload_modules)


def from_uri(uri, impl_generator=get_data_element_impls):
    """
    Create a data element instance from available plugin implementations.

    The first implementation that can resolve the URI is what is returned. If no
    implementations can resolve the URL, an ``InvalidUriError`` is raised.

    :param uri: URI to try to resolve into a DataElement instance.
    :type uri: str

    :param impl_generator: Function that returns a dictionary mapping
        implementation type names to the class type. By default this refers to
        the standard ``get_data_element_impls`` function, however this can be
        changed to refer to a custom set of classes if desired.
    :type impl_generator: () -> dict[str, type]

    :raises smqtk.exceptions.InvalidUriError: No data element implementations
        could resolve the given URI.

    :return: New data element instance providing access to the data pointed to
        by the input URI.
    :rtype: DataElement

    """
    log = logging.getLogger(__name__)
    log.debug("Trying to parse URI: '%s'", uri)

    #: :type: collections.Iterable[DataElement]
    de_type_iter = six.itervalues(impl_generator())
    inst = None
    for de_type in de_type_iter:
        try:
            inst = de_type.from_uri(uri)
        except NoUriResolutionError:
            # Expected error signaling that DataElement implementation does not
            # or cannot resolve from a URI.
            pass
        except InvalidUriError as ex:
            log.debug("Implementation '%s' failed to parse URI: %s",
                      de_type.__name__, ex.reason)
        if inst is not None:
            break
    if inst is None:
        # TODO: Assume final fallback of FileElement?
        #       Since any string could be a file?
        raise InvalidUriError(uri, "No available implementation to handle URI.")
    return inst



"""
References/Resources:
- UNC paper: http://www.cs.unc.edu/~lazebnik/publications/cvpr11_small_code.pdf
- ACM reference: http://dl.acm.org/citation.cfm?id=2191779
- GitHub with matlab implementation:
  https://github.com/willard-yuan/hashing-baseline-for-image-retrieval
"""
from collections import Sequence
from copy import deepcopy
import logging

import numpy
from six import BytesIO

from smqtk.algorithms.nn_index.lsh.functors import LshFunctor
from smqtk.representation import get_data_element_impls
from smqtk.representation.descriptor_element import elements_to_matrix
from smqtk.utils import merge_dict, plugin
from smqtk.utils.bin_utils import ProgressReporter


class ItqFunctor (LshFunctor):
    """
    LSH hash code functor implementation using Iterative Quantization (ITQ), a
    method to convert a descriptor (e.g. 4000-dim vector) to a fewer number of
    bits (e.g. 64 bits).

    The method first appeared in
    ```
    Yunchao Gong and Svetlana Lazebnik. Iterative Quantization: A Procrustes
    Approach to Learning Binary Codes. In CVPR 2011.
    ```
    It was originally implemented in Matlab by Yunchao Gong
    (yunchao@cs.unc.edu).

    IMPORTANT:
        For consistency, we treat bit vectors such that the bit at index 0 is
        considered the most significant bit (big-endian).

    """

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        default = super(ItqFunctor, cls).get_default_config()

        # Cache element parameters need to be split out into sub-configurations
        data_element_default_config = \
            plugin.make_config(get_data_element_impls())
        default['mean_vec_cache'] = data_element_default_config
        # Need to deepcopy source to prevent modifications on one sub-config
        # from reflecting in the other.
        default['rotation_cache'] = deepcopy(data_element_default_config)

        return default

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the JSON-compliant
        configuration dictionary encapsulating initialization arguments.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: ItqFunctor

        """
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        data_element_impls = get_data_element_impls()
        # Mean vector cache element.
        mean_vec_cache = None
        if config_dict['mean_vec_cache'] and \
                config_dict['mean_vec_cache']['type']:
            mean_vec_cache = plugin.from_plugin_config(
                config_dict['mean_vec_cache'], data_element_impls)
        config_dict['mean_vec_cache'] = mean_vec_cache
        # Rotation matrix cache element.
        rotation_cache = None
        if config_dict['rotation_cache'] and \
                config_dict['rotation_cache']['type']:
            rotation_cache = plugin.from_plugin_config(
                config_dict['rotation_cache'], data_element_impls)
        config_dict['rotation_cache'] = rotation_cache

        return super(ItqFunctor, cls).from_config(config_dict, False)

    def __init__(self, mean_vec_cache=None, rotation_cache=None,
                 bit_length=8, itq_iterations=50, normalize=None,
                 random_seed=None):
        """
        Initialize IQR functor.

        File Path Note
        --------------
        Model filepaths should point to numpy files and include the ".npy" file
        suffix. If not, then when files are saves, numpy will automatically
        append the ".npy" suffix (e.g. "example" -> "example.npy"). This can
        lead to not being able to use the same configuration for fitting and
        application, as the file name's as given will not be found.

        :param mean_vec_cache: Optional data element to load/store the mean
            vector when initialized and/or built. When None, this will only be
            stored in memory.
        :type mean_vec_cache: smqtk.representation.DataElement

        :param rotation_cache: Optional data element to load/store the rotation
            matrix when initialize and/or built. When None, this will only be
            stored in memory.
        :type rotation_cache: smqtk.representation.DataElement

        :param bit_length: Number of bits used to represent descriptors (hash
            code). This must be greater than 0. If given an existing
        :type bit_length: int

        :param itq_iterations: Number of iterations for the ITQ algorithm to
            perform. This must be greater than 0.
        :tyepe itq_iterations: int

        :param normalize: Normalize input vectors when fitting and generation
            hash vectors using ``numpy.linalg.norm``. This may either
            be ``None``, disabling normalization, or any valid value that
            could be passed to the ``ord`` parameter in ``numpy.linalg.norm``
            for 1D arrays. This is ``None`` by default (no normalization).

            Normalization affects the value of the mean vector and rotation
            matrix. This means that model products produced reflect the
            normalization value use when training and the same normalization
            value, like the bit_length value, must be used when loading cached
            models again for later use.
        :type normalize: None | int | float | str

        :param random_seed: Integer to use as the random number generator seed.
        :type random_seed: int

        """
        super(ItqFunctor, self).__init__()

        self.mean_vec_cache_elem = mean_vec_cache
        self.rotation_cache_elem = rotation_cache
        self.bit_length = bit_length
        self.itq_iterations = itq_iterations
        self.normalize = normalize
        self.random_seed = random_seed

        # Validate normalization parameter by trying it on a random vector
        if normalize is not None:
            self._norm_vector(numpy.random.rand(8))

        # Model components
        self.mean_vec = None
        self.rotation = None

        self.load_model()

    def _norm_vector(self, v):
        """
        Class standard array normalization. Normalized along max dimension (a=0
        for a 1D array, a=1 for a 2D array, etc.).

        If ``self.normalize`` is None, ``v`` is returned without modification.

        :param v: Vector to normalize
        :type v: numpy.ndarray

        :return: Returns the normalized version of input array ``v``.
        :rtype: numpy.ndarray

        """
        if self.normalize is not None:
            n = numpy.linalg.norm(v, self.normalize, v.ndim - 1,
                                  keepdims=True)
            # replace 0's with 1's, preventing div-by-zero
            n[n == 0.] = 1.
            return v / n

        # Normalization off
        return v

    def get_config(self):
        # If no cache elements (set to None), return default plugin configs.
        c = merge_dict(self.get_default_config(), {
            "bit_length": self.bit_length,
            "itq_iterations": self.itq_iterations,
            "normalize": self.normalize,
            "random_seed": self.random_seed,
        })
        if self.mean_vec_cache_elem:
            c['mean_vec_cache'] = \
                plugin.to_plugin_config(self.mean_vec_cache_elem)
        if self.rotation_cache_elem:
            c['rotation_cache'] = \
                plugin.to_plugin_config(self.rotation_cache_elem)
        return c

    def has_model(self):
        return (self.mean_vec is not None) and (self.rotation is not None)

    def load_model(self):
        if (self.mean_vec_cache_elem
                and not self.mean_vec_cache_elem.is_empty()
                and self.rotation_cache_elem
                and not self.rotation_cache_elem.is_empty()):
            self.mean_vec = \
                numpy.load(BytesIO(self.mean_vec_cache_elem.get_bytes()))
            self.rotation = \
                numpy.load(BytesIO(self.rotation_cache_elem.get_bytes()))

    def save_model(self):
        # Check that we have cache elements set, they are writable and that we
        # have something to save to them.
        if (self.mean_vec_cache_elem and self.rotation_cache_elem and
                self.mean_vec_cache_elem.writable() and
                self.rotation_cache_elem.writable() and
                self.mean_vec is not None and self.rotation is not None):
            b = BytesIO()
            # noinspection PyTypeChecker
            numpy.save(b, self.mean_vec)
            self.mean_vec_cache_elem.set_bytes(b.getvalue())

            b = BytesIO()
            # noinspection PyTypeChecker
            numpy.save(b, self.rotation)
            self.rotation_cache_elem.set_bytes(b.getvalue())

    def _find_itq_rotation(self, v, n_iter):
        """
        Finds a rotation of the PCA embedded data. Number of iterations must be
        greater than 0.

        This is equivalent to the ITQ function from UNC-CH's implementation.

        ``self`` is used only for logging. Otherwise this has no side effects
        on this instance.

        :param v: 2D numpy array, [n, c] shape PCA embedded data, ``n`` is the
            number of data elements and ``c`` is the code length.
        :type v: numpy.core.multiarray.ndarray

        :param n_iter: max number of iterations, 50 is usually enough
        :type n_iter: int

        :return: [b, r]
           b: 2D numpy array, [n, c] shape binary matrix,
           r: 2D numpy array, the [c, c] shape rotation matrix found by ITQ
        :rtype: numpy.core.multiarray.ndarray[bool],
            numpy.core.multiarray.ndarray[float]

        """
        # Pull num bits from PCA-projected descriptors
        bit = v.shape[1]
        # initialize with an orthogonal random rotation
        if self.random_seed is not None:
            numpy.random.seed(self.random_seed)
        r = numpy.random.randn(bit, bit)
        u11, s2, v2 = numpy.linalg.svd(r)
        r = u11[:, :bit]

        # ITQ to find optimal rotation
        self._log.debug("ITQ iterations to determine optimal rotation: %d",
                        n_iter)
        for i in range(n_iter):
            self._log.debug("ITQ iter %d", i + 1)
            # TODO: @numba.jit decorate
            z = numpy.dot(v, r)
            ux = numpy.ones(z.shape) * (-1)
            ux[z >= 0] = 1
            c = numpy.dot(ux.transpose(), v)
            ub, sigma, ua = numpy.linalg.svd(c)
            r = numpy.dot(ua, ub.transpose())

        # Make B binary matrix using final rotation matrix
        #   - original code returned B transformed by second to last rotation
        #       matrix, there by returning, de-synchronized matrices
        #   - Recomputing Z here so as to generate up-to-date B for the final
        #       rotation matrix computed.
        # TODO: Could move this step up one level and just return rotation mat?
        z = numpy.dot(v, r)
        b = numpy.zeros(z.shape, dtype=numpy.bool)
        b[z >= 0] = True

        return b, r

    def fit(self, descriptors, use_multiprocessing=True):
        """
        Fit the ITQ model given the input set of descriptors.

        :param descriptors: Iterable of ``DescriptorElement`` vectors to fit
            the model to.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param use_multiprocessing: If multiprocessing should be used, as
            opposed to threading, when collecting descriptor elements from the
            given iterable.
        :type use_multiprocessing: bool

        :raises RuntimeError: There is already a model loaded

        :return: Matrix hash codes for provided descriptors in order.
        :rtype: numpy.ndarray[bool]

        """
        if self.has_model():
            raise RuntimeError("Model components have already been loaded.")

        dbg_report_interval = 1.0
        dbg_report = self.get_logger().getEffectiveLevel() <= logging.DEBUG
        if not isinstance(descriptors, Sequence):
            self._log.info("Creating sequence from iterable")
            descriptors_l = []
            pr = ProgressReporter(self._log.debug, dbg_report_interval).start()
            for d in descriptors:
                descriptors_l.append(d)
                dbg_report and pr.increment_report()
            dbg_report and pr.report()
            descriptors = descriptors_l
        if len(descriptors[0].vector()) < self.bit_length:
            raise ValueError("Input descriptors have fewer features than "
                             "requested bit encoding. Hash codes will be "
                             "smaller than requested due to PCA decomposition "
                             "result being bound by number of features.")

        self._log.info("Creating matrix of descriptors for fitting")
        x = elements_to_matrix(descriptors, report_interval=dbg_report_interval,
                               use_multiprocessing=use_multiprocessing)
        self._log.debug("descriptor matrix shape: %s", x.shape)

        self._log.debug("Info normalizing descriptors by factor: %s",
                        self.normalize)
        x = self._norm_vector(x)

        self._log.info("Centering data")
        self.mean_vec = numpy.mean(x, axis=0)
        x -= self.mean_vec

        self._log.info("Computing PCA transformation")
        self._log.debug("-- computing covariance")
        # ``cov`` wants each row to be a feature and each column an observation
        # of those features. Thus, each column should be a descriptor vector,
        # thus we need the transpose here.
        c = numpy.cov(x.transpose())

        if True:
            # Direct translation from UNC matlab code
            # - eigen vectors are the columns of ``pc``
            self._log.debug('-- computing linalg.eig')
            l, pc = numpy.linalg.eig(c)
            self._log.debug('-- ordering eigen vectors by descending eigen '
                            'value')
        else:
            # Harry translation -- Uses singular values / vectors, not eigen
            # - singular vectors are the columns of pc
            self._log.debug('-- computing linalg.svd')
            pc, l, _ = numpy.linalg.svd(c)
            self._log.debug('-- ordering singular vectors by descending '
                            'singular value')

        # Same ordering method for both eig/svd sources.
        l_pc_ordered = sorted(zip(l, pc.transpose()), key=lambda _p: _p[0],
                              reverse=True)

        self._log.debug("-- top vector extraction")
        # Only keep the top ``bit_length`` vectors after ordering by descending
        # value magnitude.
        # - Transposing vectors back to column-vectors.
        pc_top = numpy.array([p[1] for p in l_pc_ordered[:self.bit_length]])\
            .transpose()
        self._log.debug("-- project centered data by PC matrix")
        v = numpy.dot(x, pc_top)

        self._log.info("Performing ITQ to find optimal rotation")
        c, self.rotation = self._find_itq_rotation(v, self.itq_iterations)
        # De-adjust rotation with PC vector
        self.rotation = numpy.dot(pc_top, self.rotation)

        self.save_model()

        return c

    def get_hash(self, descriptor):
        """
        Get the locality-sensitive hash code for the input descriptor.

        :param descriptor: Descriptor vector we should generate the hash of.
        :type descriptor: numpy.ndarray[float]

        :return: Generated bit-vector as a numpy array of booleans.
        :rtype: numpy.ndarray[bool]

        """
        z = numpy.dot(self._norm_vector(descriptor) - self.mean_vec,
                      self.rotation)
        b = numpy.zeros(z.shape, dtype=bool)
        b[z >= 0] = True
        return b



from smqtk.algorithms.classifier import Classifier, SupervisedClassifier


STUB_CLASSIFIER_MOD_PATH = __name__


class DummyClassifier (Classifier):
    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this
        class's ``from_config`` method to produce an instance with identical
        configuration.

        In the common case, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        # No constructor, no config
        return {}

    @classmethod
    def is_usable(cls):
        """
        Check whether this class is available for use.

        Since certain plugin implementations may require additional
        dependencies that may not yet be available on the system, this method
        should check for those dependencies and return a boolean saying if the
        implementation is usable.

        NOTES:
            - This should be a class method
            - When an implementation is deemed not usable, this should emit a
                warning detailing why the implementation is not available for
                use.

        :return: Boolean determination of whether this implementation is
                 usable.
        :rtype: bool

        """
        return True

    def _classify(self, d):
        """
        Internal method that constructs the label-to-confidence map (dict) for
        a given DescriptorElement.

        The passed descriptor element is guaranteed to have a vector to
        extract. It is not extracted yet due to the philosophy of waiting
        until the vector is immediately needed. This moment is thus determined
        by the implementing algorithm.

        :param d: DescriptorElement containing the vector to classify.
        :type d: smqtk.representation.DescriptorElement

        :raises RuntimeError: Could not perform classification for some reason
            (see message in raised exception).

        :return: Dictionary mapping trained labels to classification
            confidence values
        :rtype: dict[collections.Hashable, float]

        """
        return {
            'negative': 0.5,
            'positive': 0.5,
        }

    def get_labels(self):
        """
        Get the sequence of class labels that this classifier can classify
        descriptors into. This includes the negative label.

        :return: Sequence of possible classifier labels.
        :rtype: collections.Sequence[collections.Hashable]

        :raises RuntimeError: No model loaded.

        """
        return ['negative', 'positive']


class DummySupervisedClassifier (SupervisedClassifier):
    """
    Supervise classifier stub implementation.
    """

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        pass

    def has_model(self):
        pass

    def _train(self, class_examples, **extra_params):
        pass

    def get_labels(self):
        pass

    def _classify(self, d):
        pass



"""
Data abstraction interface for general key-value storage.
"""
import abc
import collections
import os

from smqtk.exceptions import ReadOnlyError
from smqtk.representation import SmqtkRepresentation
from smqtk.utils.plugin import Pluggable


NO_DEFAULT_VALUE = type("KeyValueStoreNoDefaultValueType", (object,), {})()


class KeyValueStore (SmqtkRepresentation, Pluggable):
    """
    Interface for general key/value storage.

    Implementations may impose restrictions on what types keys or values may be
    due to backend used.

    Data access and manipulation should be thread-safe.
    """

    # Mutable storage container is not hashable.
    __hash__ = None

    def __len__(self):
        return self.count()

    def __contains__(self, item):
        return self.has(item)

    def __getitem__(self, item):
        return self.get(item)

    @abc.abstractmethod
    def __repr__(self):
        """
        Return representative string for this class.

        *NOTE:* **This abstract super-method returns a template string to add
        to sub-class specific information to. The returned string should be
        formatted using the ``%`` operator and expects a single string
        argument.**

        :return: Representative string for this class.
        :rtype: str

        """
        return '<' + self.__class__.__name__ + " %s>"

    @abc.abstractmethod
    def count(self):
        """
        :return: The number of key-value relationships in this store.
        :rtype: int | long
        """

    @abc.abstractmethod
    def keys(self):
        """
        :return: Iterator over keys in this store.
        :rtype: collections.Iterator[collections.Hashable]
        """

    def values(self):
        """
        :return: Iterator over values in this store. Values are not guaranteed
            to be in any particular order.
        :rtype: collections.Iterator[object]
        """
        for k in self.keys():
            yield self.get(k)

    @abc.abstractmethod
    def is_read_only(self):
        """
        :return: True if this instance is read-only and False if it is not.
        :rtype: bool
        """

    @abc.abstractmethod
    def has(self, key):
        """
        Check if this store has a value for the given key.

        :param key: Key to check for a value for.
        :type key: collections.Hashable

        :return: If this store has a value for the given key.
        :rtype: bool

        """

    @abc.abstractmethod
    def add(self, key, value):
        """
        Add a key-value pair to this store.

        *NOTE:* **Implementing sub-classes should call this super-method. This
        super method should not be considered a critical section for thread
        safety unless ``is_read_only`` is not thread-safe.**

        :param key: Key for the value. Must be hashable.
        :type key: collections.Hashable

        :param value: Python object to store.
        :type value: object

        :raises ReadOnlyError: If this instance is marked as read-only.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot add to read-only instance %s." % self)

    @abc.abstractmethod
    def add_many(self, d):
        """
        Add multiple key-value pairs at a time into this store as represented
        in the provided dictionary `d`.

        :param d: Dictionary of key-value pairs to add to this store.
        :type d: dict[collections.Hashable, object]

        :raises ReadOnlyError: If this instance is marked as read-only.

        :return: Self.
        :rtype: KeyValueStore

        """
        # Input keys must already be hashable because they're in a dictionary.
        if self.is_read_only():
            raise ReadOnlyError("Cannot add to read-only instance %s." % self)

    @abc.abstractmethod
    def remove(self, key):
        """
        Remove a single key-value entry.

        :param key: Key to remove.
        :type key: collections.Hashable

        :raises ReadOnlyError: If this instance is marked as read-only.
        :raises KeyError: The given key is not present in this store and no
            default value given.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot remove from read-only instance %s."
                                % self)

    @abc.abstractmethod
    def remove_many(self, keys):
        """
        Remove multiple keys and associated values.

        :param keys: Iterable of keys to remove.  If this is empty this method
            does nothing.
        :type keys: collections.Iterable[collections.Hashable]

        :raises ReadOnlyError: If this instance is marked as read-only.
        :raises KeyError: The given key is not present in this store and no
            default value given.  The store is not modified if any key is
            invalid.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot remove from read-only instance %s."
                                % self)

    @abc.abstractmethod
    def get(self, key, default=NO_DEFAULT_VALUE):
        """
        Get the value for the given key.

        *NOTE:* **Implementing sub-classes are responsible for raising a
        ``KeyError`` where appropriate.**

        :param key: Key to get the value of.
        :type key: collections.Hashable

        :param default: Optional default value if the given key is not present
            in this store. This may be any value except for the
            ``NO_DEFAULT_VALUE`` constant (custom anonymous class instance).
        :type default: object

        :raises KeyError: The given key is not present in this store and no
            default value given.

        :return: Deserialized python object stored for the given key.
        :rtype: object

        """

    # TODO: get_many(self, keys, default=NO_DEFAULT_VALUE)

    @abc.abstractmethod
    def clear(self):
        """
        Clear this key-value store.

        *NOTE:* **Implementing sub-classes should call this super-method. This
        super method should not be considered a critical section for thread
        safety.**

        :raises ReadOnlyError: If this instance is marked as read-only.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot clear a read-only %s instance."
                                % self.__class__.__name__)


def get_key_value_store_impls(reload_modules=False):
    """
    Discover and return discovered ``KeyValueStore`` classes. Keys in the
    returned map are the names of the discovered classes, and the paired values
    are the actual class type objects.

    We search for implementation classes in:
        - modules next to this file this function is defined in (ones that begin
          with an alphanumeric character),
        - python modules listed in the environment variable
          ``KEY_VALUE_STORE_PATH``
            - This variable should contain a sequence of python module
              specifications, separated by the platform specific PATH separator
              character (``;`` for Windows, ``:`` for unix)

    Within a module we first look for a helper variable by the name
    ``KEY_VALUE_STORE_CLASS``, which can either be a single class object or
    an iterable of class objects, to be specifically exported. If the variable
    is set to None, we skip that module and do not import anything. If the
    variable is not present, we look at attributes defined in that module for
    classes that descend from the given base class type. If none of the above
    are found, or if an exception occurs, the module is skipped.

    :param reload_modules: Explicitly reload discovered modules from source.
    :type reload_modules: bool

    :return: Map of discovered class object of type ``KeyValueStore``
        whose keys are the string names of the classes.
    :rtype: dict[str, type]

    """
    from smqtk.utils.plugin import get_plugins
    this_dir = os.path.abspath(os.path.dirname(__file__))
    env_var = "KEY_VALUE_STORE_PATH"
    helper_var = "KEY_VALUE_STORE_CLASS"
    return get_plugins(__name__, this_dir, env_var, helper_var,
                       KeyValueStore, reload_modules=reload_modules)
